"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
import inspect

from nagini_translation.lib.constants import (
    BOOL_TYPE,
    BOXED_PRIMITIVES,
    BYTES_TYPE,
    CHECK_DEFINED_FUNC,
    DICT_TYPE,
    END_LABEL,
    FLOAT_TYPE,
    FUNCTION_DOMAIN_NAME,
    GET_METHOD_FUNC,
    GLOBAL_VAR_FIELD,
    INT_TYPE,
    LIST_TYPE,
    METHOD_ID_DOMAIN,
    OPERATOR_FUNCTIONS,
    PRIMITIVE_INT_TYPE,
    SET_TYPE,
    STRING_TYPE,
    THREAD_DOMAIN,
    TUPLE_TYPE,
)
from nagini_translation.lib.errors import rules
from nagini_translation.lib.program_nodes import (
    GenericType,
    OptionalType,
    PythonClass,
    PythonField,
    PythonGlobalVar,
    PythonIOExistentialVar,
    PythonMethod,
    PythonModule,
    PythonTryBlock,
    PythonType,
    PythonVar,
    UnionType,
    toposort_classes,
    chain_cond_exp,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Position,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.util import (
    construct_lambda_prefix,
    get_func_name,
    get_surrounding_try_blocks,
    InvalidProgramException,
    join_three_expressions,
    join_expressions,
    UnsupportedException,
)
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator
from typing import List, Optional, Tuple, Union


# Maps function names to bools; caches which functions take an additonal argument
# encoding if impure assertions are allowed or not.
TAKES_IMPURE_ARGS = {}


class ExpressionTranslator(CommonTranslator):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # TODO: Update all code to use this flag.
        self._is_expression = False
        self._target_type = None
        self._as_read = False
        self._primitive_operations = {
            ast.Add: self.viper.Add,
            ast.Sub: self.viper.Sub,
            ast.Mult: self.viper.Mul,
            ast.FloorDiv: self.viper.Div,
            ast.Mod: self.viper.Mod,
            ast.Is: self.viper.EqCmp,
            ast.Eq: self.viper.EqCmp,
            ast.NotEq: self.viper.NeCmp,
            ast.IsNot: self.viper.NeCmp,
            ast.Lt: self.viper.LtCmp,
            ast.LtE: self.viper.LeCmp,
            ast.Gt: self.viper.GtCmp,
            ast.GtE: self.viper.GeCmp,
        }

    def translate_expr(self, node: ast.AST, ctx: Context,
                       target_type = None,
                       impure: bool = False,
                       as_read: bool = False) -> StmtsAndExpr:
        """
        Generic visitor function for translating an expression.

        :param target_type:
            The Silver type this expression should be translated to, defaults
            to Ref if no arguments is provided.
        :param impure:
            Indicates if ``node`` may be translated to an impure assertion. If False,
            translating a predicate or field permission will result in an
            InvalidProgramException.
        :param as_read:
            Indicates if ``node`` should be translated as if it was being read (as
            opposed to assigned to) independently of the actual context of the AST node.
        """
        old_as_read = self._as_read
        self._as_read = as_read
        if not target_type:
            target_type = self.viper.Ref
        old_target = self._target_type
        self._target_type = target_type
        stmt, result = self._translate_only(node, ctx, impure)

        if target_type != result.typ():
            result = self.convert_to_type(result, target_type, ctx, node)

        self._target_type = old_target
        self._as_read = old_as_read
        return stmt, result

    def _translate_only(self, node: ast.AST, ctx: Context,
                        impure = False) -> StmtsAndExpr:
        """
        Translates an expression, but does so without changing the expression's
        type in any way.
        """
        method = 'translate_' + node.__class__.__name__
        visitor = getattr(self, method, self.translate_generic)
        if method in TAKES_IMPURE_ARGS:
            impure_arg = TAKES_IMPURE_ARGS[method]
        else:
            sig = inspect.signature(visitor)
            impure_arg = len(sig.parameters) > 2
            TAKES_IMPURE_ARGS[method] = impure_arg
        if impure_arg:
            stmt, result = visitor(node, ctx, impure)
        else:
            stmt, result = visitor(node, ctx)
        return stmt, result

    def translate_Return(self, node: ast.Return, ctx: Context,
                         impure=False) -> StmtsAndExpr:
        return self.translate_expr(node.value, ctx, impure=impure)

    def translate_ListComp(self, node: ast.ListComp, ctx: Context) -> StmtsAndExpr:
        if len(node.generators) != 1:
            raise UnsupportedException(node, 'Multiple generators in list comprehension.')
        if node.generators[0].ifs:
            raise UnsupportedException(node, 'Filter in list comprehension.')
        list_class = ctx.module.global_module.classes['list']
        name = construct_lambda_prefix(node.lineno, node.col_offset)
        target = node.generators[0].target
        local_name = name + '$' + target.id
        element_var = ctx.actual_function.special_vars[local_name]
        ctx.set_alias(target.id, element_var)
        body_stmt, body = self.translate_expr(node.elt, ctx)
        result_type = self.get_type(node.elt, ctx)
        list_type = GenericType(list_class, [result_type])
        if body_stmt:
            raise InvalidProgramException(node, 'impure.list.comprehension.body')
        ctx.remove_alias(target.id)
        result_var = ctx.actual_function.create_variable('listcomp', list_type,
                                                         self.translator)
        stmt = self._create_list_comp_inhale(result_var, list_type, element_var,
                                             body, node, ctx)
        return stmt, result_var.ref()

    def _create_list_comp_inhale(self, result_var: PythonVar, list_type: PythonType,
                                 element_var: PythonVar, body: Expr, node: ast.ListComp,
                                 ctx: Context) -> List[Stmt]:
        # Create an inhale of the following form:
        # Inhale issubtype(typeof(result), list(result_type)) && acc(result.list_acc) &&
        # len(result) == |iter.__sil_seq__()| &&
        # forall 0 <= i < len(result) :: result.list_acc[i] ==
        #                                (let e == iter.__sil_seq__()[i] in body)
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        list_class = ctx.module.global_module.classes['list']
        type_check = self.type_check(result_var.ref(), list_type, position, ctx, False)
        list_acc_field = self.viper.Field('list_acc', self.viper.SeqType(self.viper.Ref),
                                          position, info)
        field_acc = self.viper.FieldAccess(result_var.ref(), list_acc_field, position,
                                           info)
        acc_pred = self.viper.FieldAccessPredicate(field_acc,
                                                   self.viper.FullPerm(position, info),
                                                   position, info)
        type_and_perm = self.viper.And(type_check, acc_pred, position, info)
        result_len = self.get_function_call(list_class, '__len__', [result_var.ref()],
                                            [None], node, ctx)
        iter_stmt, iter = self.translate_expr(node.generators[0].iter, ctx)
        iter_type = self.get_type(node.generators[0].iter, ctx)
        sil_seq = self.get_sequence(iter_type.python_class, iter, None, node, ctx,
                                    position)
        seq_len = self.viper.SeqLength(sil_seq, position, info)
        len_equal = self.viper.EqCmp(self.to_int(result_len, ctx), seq_len, position,
                                     info)
        int_class = ctx.module.global_module.classes[PRIMITIVE_INT_TYPE]
        index_var = ctx.actual_function.create_variable('i', int_class, self.translator,
                                                        False)
        index_positive = self.viper.GeCmp(index_var.ref(),
                                          self.viper.IntLit(0, position, info),
                                          position, info)
        index_lt_len = self.viper.LtCmp(index_var.ref(), seq_len, position, info)
        index_in_bounds = self.viper.And(index_positive, index_lt_len, position, info)
        value_i = self.viper.SeqIndex(field_acc, index_var.ref(), position, info)
        sil_seq_i = self.get_function_call(iter_type.cls, '__getitem__',
                                           [iter, index_var.ref()], [None, None],
                                           node, ctx)

        let = self.viper.Let(element_var.decl, sil_seq_i, body, position, info)
        value = self.viper.EqCmp(value_i, let, position, info)
        trigger = self.viper.Trigger([value_i], position, info)
        value_in_bounds = self.viper.Implies(index_in_bounds, value, position, info)
        values = self.viper.Forall([index_var.decl], [trigger], value_in_bounds, position,
                                   info)
        contents = self.viper.And(len_equal, values, position, info)
        all = self.viper.And(type_and_perm, contents, position, info)
        inhale = self.viper.Inhale(all, position, info)
        return iter_stmt + [inhale]

    def translate_Num(self, node: ast.Num, ctx: Context) -> StmtsAndExpr:
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        if isinstance(node.n, int):
            lit = self.viper.IntLit(node.n, pos, info)
            int_class = ctx.module.global_module.classes[PRIMITIVE_INT_TYPE]
            boxed_lit = self.get_function_call(int_class, '__box__', [lit],
                                               [None], node, ctx)
            return [], boxed_lit
        if isinstance(node.n, float):
            float_class = ctx.module.global_module.classes[FLOAT_TYPE]
            index_lit = self.viper.IntLit(ctx.get_fresh_int(), pos, info)
            float_val = self.get_function_call(float_class, '__create__', [index_lit],
                                               [None], node, ctx, pos)
            return [], float_val
        raise UnsupportedException(node, 'Unsupported number literal')

    def translate_Dict(self, node: ast.Dict, ctx: Context) -> StmtsAndExpr:
        args = []
        res_var = ctx.current_function.create_variable('dict',
            ctx.module.global_module.classes[DICT_TYPE], self.translator)
        dict_class = ctx.module.global_module.classes[DICT_TYPE]
        arg_types = []
        constr_call = self.get_method_call(dict_class, '__init__', [],
                                           [], [res_var.ref()], node, ctx)
        stmt = constr_call
        # Inhale the type of the newly created dict (including type arguments)
        dict_type = self.get_type(node, ctx)
        position = self.to_position(node, ctx)
        stmt.append(self.viper.Inhale(self.type_check(res_var.ref(node, ctx),
                                                      dict_type, position, ctx),
                                      position, self.no_info(ctx)))
        for key, val in zip(node.keys, node.values):
            key_stmt, key_val = self.translate_expr(key, ctx)
            key_type = self.get_type(key, ctx)
            val_stmt, val_val = self.translate_expr(val, ctx)
            val_type = self.get_type(val, ctx)
            args = [res_var.ref(), key_val, val_val]
            arg_types = [None, key_type, val_type]
            append_call = self.get_method_call(dict_class, '__setitem__', args,
                                               arg_types, [], node, ctx)
            stmt += key_stmt + val_stmt + append_call
        return stmt, res_var.ref(node, ctx)

    def translate_Set(self, node: ast.Set, ctx: Context) -> StmtsAndExpr:
        set_class = ctx.module.global_module.classes[SET_TYPE]
        res_var = ctx.current_function.create_variable(SET_TYPE,
            set_class, self.translator)
        constr_call = self.get_method_call(set_class, '__init__', [], [],
                                           [res_var.ref()], node, ctx)
        stmt = constr_call
        # Inhale the type of the newly created set (including type arguments)
        set_type = self.get_type(node, ctx)
        content_type = set_type.type_args[0]
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        stmt.append(self.viper.Inhale(self.type_check(res_var.ref(node, ctx),
                                                      set_type, position, ctx),
                                      position, self.no_info(ctx)))
        elements = []
        assert_pos = self.to_position(node, ctx, error_string='element type is correct')
        for element in node.elts:
            el_stmt, el = self.translate_expr(element, ctx)
            check = self.viper.Assert(self.type_check(el, content_type, assert_pos, ctx,
                                                      False), assert_pos, info)
            stmt += el_stmt + [check]
            elements.append(el)
        result = res_var.ref(node, ctx)
        if elements:
            content_field = self.viper.Field('set_acc', self.viper.SetType(self.viper.Ref),
                                             position, info)
            field_acc = self.viper.FieldAccess(result, content_field, position, info)
            content_seq = self.viper.ExplicitSet(elements, position, info)
            stmt.append(self.viper.FieldAssign(field_acc, content_seq, position, info))
        return stmt, result

    def translate_List(self, node: ast.List, ctx: Context) -> StmtsAndExpr:
        list_class = ctx.module.global_module.classes[LIST_TYPE]
        res_var = ctx.current_function.create_variable(LIST_TYPE,
            list_class, self.translator)
        constr_call = self.get_method_call(list_class, '__init__', [], [],
                                           [res_var.ref()], node, ctx)
        stmt = constr_call
        # Inhale the type of the newly created list (including type arguments)
        list_type = self.get_type(node, ctx)
        content_type = list_type.type_args[0]
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        stmt.append(self.viper.Inhale(self.type_check(res_var.ref(node, ctx),
                                                      list_type, position, ctx),
                                      position, self.no_info(ctx)))
        elements = []
        assert_pos = self.to_position(node, ctx, error_string='element type is correct')
        for element in node.elts:
            el_stmt, el = self.translate_expr(element, ctx)
            check = self.viper.Assert(self.type_check(el, content_type, assert_pos, ctx,
                                                      False), assert_pos, info)
            stmt += el_stmt + [check]
            elements.append(el)

        result = res_var.ref(node, ctx)
        if elements:
            content_field = self.viper.Field('list_acc', self.viper.SeqType(self.viper.Ref),
                                             position, info)
            field_acc = self.viper.FieldAccess(result, content_field, position, info)
            content_seq = self.viper.ExplicitSeq(elements, position, info)
            stmt.append(self.viper.FieldAssign(field_acc, content_seq, position, info))
        return stmt, result

    def translate_Str(self, node: ast.Str, ctx: Context) -> StmtsAndExpr:
        return [], self.translate_string(node.s, node, ctx)

    def translate_string(self, s: str, node: ast.AST, ctx: Context) -> Expr:
        length = len(s)
        length_arg = self.viper.IntLit(length, self.no_position(ctx),
                                       self.no_info(ctx))
        val_arg = self.viper.IntLit(self._get_string_value(s),
                                    self.no_position(ctx), self.no_info(ctx))
        args = [length_arg, val_arg]
        arg_types = [None, None]
        str_type = ctx.module.global_module.classes[STRING_TYPE]
        func_name = '__create__'
        call = self.get_function_call(str_type, func_name, args, arg_types,
                                      node, ctx)
        return call


    def translate_Bytes(self, node: ast.Bytes, ctx: Context) -> StmtsAndExpr:
        elems = []
        for c in node.s:
            lit = self.viper.IntLit(c, self.to_position(node, ctx),
                                    self.no_info(ctx))
            elems.append(self.to_ref(lit, ctx))
        if elems:
            seq = self.viper.ExplicitSeq(elems, self.to_position(node, ctx),
                                         self.no_info(ctx))
        else:
            seq = self.viper.EmptySeq(self.viper.Ref,
                                      self.to_position(node, ctx),
                                      self.no_info(ctx))
        bytes_class = ctx.module.global_module.classes[BYTES_TYPE]
        args = [seq, self.get_fresh_int_lit(ctx)]
        result = self.get_function_call(bytes_class, '__create__', args,
                                        [None, None], node, ctx)
        return [], result

    def translate_Tuple(self, node: ast.Tuple, ctx: Context) -> StmtsAndExpr:
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        stmts = []
        vals = []
        val_types = []
        for el in node.elts:
            el_stmt, el_val = self.translate_expr(el, ctx)
            stmts += el_stmt
            vals.append(el_val)
            val_types.append(self.get_type(el, ctx))
        call = self.create_tuple(vals, val_types, node, ctx)
        return stmts, call

    def create_tuple(self, vals: List[Expr], val_types: List[PythonType],
                     node: ast.AST, ctx: Context) -> Expr:
        """
        Creates a tuple containing the given values.
        """
        tuple_class = ctx.module.global_module.classes[TUPLE_TYPE]
        func_name = '__create' + str(len(vals)) + '__'
        types = [self.get_tuple_type_arg(v, t, node, ctx)
                 for (t, v) in zip(val_types, vals)]
        args = vals + types
        # Also add a running integer s.t. other tuples with same contents are
        # not reference-identical (except for empty tuples).
        if args:
            args.append(self.get_fresh_int_lit(ctx))
        arg_types = [None] * len(args)
        call = self.get_function_call(tuple_class, func_name, args, arg_types,
                                      node, ctx)
        return call

    def translate_Subscript(self, node: ast.Subscript,
                            ctx: Context) -> StmtsAndExpr:
        target_type = self.get_type(node.value, ctx)
        target_stmt, target = self.translate_expr(node.value, ctx,
                                                  target_type=self.viper.Ref)
        if not isinstance(node.slice, ast.Index):
            return self._translate_slice_subscript(node, target, target_type,
                                                   target_stmt, ctx)

        index_stmt, index = self.translate_expr(node.slice.value, ctx,
                                                target_type=self.viper.Ref)
        index_type = self.get_type(node.slice.value, ctx)
        args = [target, index]
        arg_types = [target_type, index_type]
        call_stmt, call = self.get_func_or_method_call(target_type, '__getitem__', args,
                                                       arg_types, node, ctx)
        return target_stmt + index_stmt + call_stmt, call

    def _translate_slice_subscript(self, node: ast.Subscript, target: Expr,
                                   target_type: PythonType,
                                   target_stmt: List[Stmt],
                                   ctx: Context) -> StmtsAndExpr:
        if node.slice.step:
            raise UnsupportedException(node, 'slice step')
        slice_class = ctx.module.global_module.classes['slice']
        null = self.viper.NullLit(self.no_position(ctx), self.no_info(ctx))
        start = stop = null
        start_stmt = stop_stmt = []
        if node.slice.lower:
            start_stmt, start = self.translate_expr(node.slice.lower, ctx)
        if node.slice.upper:
            stop_stmt, stop = self.translate_expr(node.slice.upper, ctx)
        slice = self.get_function_call(slice_class, '__create__',
                                       [start, stop], [None, None],
                                       node.slice, ctx)
        args = [target, slice]
        stmt = target_stmt + start_stmt + stop_stmt
        getitem = target_type.get_func_or_method('__getitem_slice__')
        if not getitem.pure:
            result_var = ctx.actual_function.create_variable(
                'slice_res', target_type, self.translator)
            call = self.get_method_call(target_type, '__getitem_slice__',
                                        args, [None, None],
                                        [result_var.ref()], node, ctx)
            stmt += call
            return stmt, result_var.ref()
        else:
            call = self.get_function_call(target_type, '__getitem_slice__',
                                          args, [None, None], node, ctx)
            return stmt, call

    def create_exception_catchers(self, var: PythonVar,
                                  try_blocks: List[PythonTryBlock],
                                  call: ast.Call, ctx: Context) -> List[Stmt]:
        """
        Creates the code for catching an exception, i.e. redirecting control
        flow to the handlers, to a finally block, or giving the exception to
        the caller function.
        """
        if isinstance(var, PythonVar):
            var = var.ref()
        cases = []
        position = self.to_position(call, ctx)
        err_var = ctx.error_var.ref()
        relevant_try_blocks = get_surrounding_try_blocks(try_blocks, call)
        goto_finally = self._create_goto_finally(relevant_try_blocks, ctx)
        if goto_finally:
            uncaught_option = goto_finally
        else:
            end_label = ctx.get_label_name(END_LABEL)
            goto_end = self.viper.Goto(end_label, position, self.no_info(ctx))
            if ctx.actual_function.declared_exceptions:
                assignerror = self.viper.LocalVarAssign(err_var, var, position,
                                                        self.no_info(ctx))
                uncaught_option = self.translate_block([assignerror, goto_end],
                                                       position,
                                                       self.no_info(ctx))
            else:
                error_string = '"method raises no exceptions"'
                error_pos = self.to_position(call, ctx, error_string)
                exhale_false = self.viper.Exhale(
                    self.viper.FalseLit(error_pos, self.no_info(ctx)),
                    error_pos,
                    self.no_info(ctx))
                uncaught_option = self.translate_block([exhale_false, goto_end],
                                                       position,
                                                       self.no_info(ctx))

        for block in relevant_try_blocks:
            for handler in block.handlers:
                condition = self.type_check(var, handler.exception,
                                            self.to_position(handler.node, ctx),
                                            ctx, inhale_exhale=False)
                label_name = ctx.get_label_name(handler.name)
                goto = self.viper.Goto(label_name,
                                       self.to_position(handler.node, ctx),
                                       self.no_info(ctx))
                cases.insert(0, (condition, goto))
            if block.finally_block:
                break

        result = None
        for cond, goto in cases:
            if result is None:
                result = self.viper.If(cond, goto,
                                       uncaught_option,
                                       self.to_position(handler.node, ctx),
                                       self.no_info(ctx))
            else:
                result = self.viper.If(cond, goto, result,
                                       self.to_position(handler.node, ctx),
                                       self.no_info(ctx))
        if result is None:
            error_case = uncaught_option
        else:
            error_case = result
        errnotnull = self.viper.NeCmp(var,
                                      self.viper.NullLit(self.no_position(ctx),
                                                         self.no_info(ctx)),
                                      position, self.no_info(ctx))
        emptyblock = self.translate_block([], self.no_position(ctx),
                                          self.no_info(ctx))
        errcheck = self.viper.If(errnotnull, error_case, emptyblock,
                                 position,
                                 self.no_info(ctx))
        return [errcheck]

    def _create_goto_finally(self, tries: List[PythonTryBlock],
                             ctx: Context) -> Optional[Stmt]:
        """
        If any of the blocks in tries has a finally-block, creates and
        returns the statements to jump there.
        """
        for try_ in tries:
            if try_.finally_block or try_.with_item:
                # Propagate return value
                var_next = try_.get_finally_var(self.translator)
                if var_next.sil_name in ctx.var_aliases:
                    var_next = ctx.var_aliases[var_next.sil_name]
                number_two = self.viper.IntLit(2, self.no_position(ctx),
                                               self.no_info(ctx))
                next_assign = self.viper.LocalVarAssign(var_next.ref(),
                                                        number_two,
                                                        self.no_position(ctx),
                                                        self.no_info(ctx))
                # Goto finally block
                label_name = ctx.get_label_name(try_.finally_name)
                goto_next = self.viper.Goto(label_name,
                                            self.no_position(ctx),
                                            self.no_info(ctx))
                return_block = [next_assign, goto_next]
                result = self.translate_block(return_block,
                                              self.no_position(ctx),
                                              self.no_info(ctx))
                return result
        return None

    def translate_Expr(self, node: ast.Expr, ctx: Context) -> StmtsAndExpr:
        return self.translate_expr(node.value, ctx)

    def translate_global_var_reference(self, target: PythonGlobalVar, node: ast.AST,
                                       ctx: Context) -> Expr:
        """
        Translates a reference to the given PythonGlobalVar (which could represent either
        an actual global variable or a static field).
        """
        position = self.to_position(node, ctx)
        if target.cls:
            # This is a static field
            field_func = self.translate_static_field_access(target,
                                                            ctx.current_class, node,
                                                            ctx)
            return [], field_func
        var = target
        type = self.translate_type(var.type, ctx)

        if not target.is_final:
            func_app = self.viper.FuncApp(var.sil_name, [], position,
                                          self.no_info(ctx), self.viper.Ref, [])
            global_field = self.viper.Field(GLOBAL_VAR_FIELD, type, position,
                                            self.no_info(ctx))
            res = self.viper.FieldAccess(func_app, global_field, position,
                                         self.no_info(ctx))
        else:
            res = self.viper.FuncApp(var.sil_name, [], position,
                                     self.no_info(ctx), type, [])
        if node and not isinstance(node.ctx, ast.Store) or self._as_read:
            if self.is_main_method(ctx):
                res = self.wrap_global_defined_check(res, target, ctx.module, node,
                                                     ctx)
            elif ctx.current_function:
                ctx.current_function.call_deps.add((node, target,
                                                    ctx.current_function.module))

        return [], res

    def translate_Name(self, node: ast.Name, ctx: Context) -> StmtsAndExpr:
        target = self.get_target(node, ctx)
        if isinstance(target, PythonGlobalVar):
            return self.translate_global_var_reference(target, node, ctx)
        elif isinstance(target, PythonMethod):
            func = self.viper.DomainFuncApp(target.func_constant, [],
                                            self.viper.function_domain_type(),
                                            self.to_position(node, ctx), self.no_info(ctx),
                                            FUNCTION_DOMAIN_NAME)
            return [], func
        else:
            if isinstance(target, PythonType):
                return [], self.type_factory.translate_type_literal(target,
                    self.to_position(node, ctx), ctx)
            if node.id in ctx.var_aliases:
                var = ctx.var_aliases[node.id]
            else:
                var = ctx.actual_function.get_variable(node.id)
            if (isinstance(var, PythonIOExistentialVar) and
                    not var.is_defined()):
                raise InvalidProgramException(
                    node, 'io_existential_var.use_of_undefined')

            if (isinstance(ctx.actual_function, PythonMethod) and
                    not (ctx.actual_function.pure or ctx.actual_function.predicate) and
                    not isinstance(node.ctx, ast.Store) and
                    self.is_local_variable(var, ctx)):
                result = self.wrap_definedness_check(var.ref(node, ctx), var, node, ctx)
            else:
                result = var.ref(node, ctx)
            return [], result

    def wrap_definedness_check(self, e: Expr, var: PythonVar, node: ast.AST,
                               ctx: Context) -> Expr:
        """
        Create an access to the given variable, wrapped into a function call which checks
        if the variable has been defined.
        """
        pos = self.to_position(node, ctx, rules=rules.LOCAL_VARIABLE_NOT_DEFINED)
        info = self.no_info(ctx)
        id_param_decl = self.viper.LocalVarDecl('id', self.viper.Int, pos, info)
        var_param_decl = self.viper.LocalVarDecl('val', self.viper.Ref, pos, info)
        id = self.viper.IntLit(self._get_string_value(var.sil_name), pos, info)
        return self.viper.FuncApp(CHECK_DEFINED_FUNC, [e, id], pos, info,
                                  self.viper.Ref, [var_param_decl, id_param_decl])

    def _lookup_field(self, node: ast.Attribute, ctx: Context) -> PythonField:
        """
        Returns the PythonField for a given ast.Attribute node.
        """
        recv = self.get_type(node.value, ctx)
        if recv.name == 'type':
            recv = recv.type_args[0]
        field = recv.get_field(node.attr)
        if not field:
            if (isinstance(recv, PythonType) and
                    recv.python_class.get_static_field(node.attr)):
                var = recv.python_class.static_fields[node.attr]
                return var
            raise InvalidProgramException(node, 'field.nonexistent')
        if isinstance(field, PythonField):
            field = field.actual_field
            if field.is_mangled() and (field.cls is not ctx.current_class and
                                       field.cls is not ctx.actual_function.cls):
                raise InvalidProgramException(node, 'private.field.access')

        return field

    def translate_static_field_access(self, field: PythonGlobalVar,
                                      receiver: Union[Expr, PythonType],
                                      node: ast.AST, ctx: Context) -> Expr:
        """
        Translates an access to the given field via the given receiver. The
        receiver can either be the type literal via which the field is
        accessed, an expression of type 'type' (e.g. cls in classmethods) or
        a normal object.
        """
        position = self.to_position(node, ctx)
        while field.overrides:
            field = field.overrides
        field_type = self.translate_type(field.type, ctx)

        if isinstance(receiver, PythonType):
            type_arg = self.type_factory.translate_type_literal(receiver,
                                                                position, ctx)
        else:
            if receiver.typ() != self.type_factory.type_type():
                # Normal object, get its type.
                type_arg = self.type_factory.typeof(receiver, ctx)
            else:
                # Type expression, use it directly.
                type_arg = receiver
        info = self.no_info(ctx)
        param = self.viper.LocalVarDecl('receiver',
                                        self.type_factory.type_type(), position,
                                        info)
        res = self.viper.FuncApp(field.sil_name, [type_arg], position, info,
                                 field_type, [param])
        if not field.is_final:
            global_field = self.viper.Field(GLOBAL_VAR_FIELD, field_type, position,
                                            self.no_info(ctx))
            res = self.viper.FieldAccess(res, global_field, position,
                                         self.no_info(ctx))
        if isinstance(node, ast.Attribute):
            node = node.value
        if isinstance(receiver, PythonType) and not isinstance(node.ctx, ast.Store):
            if self.is_main_method(ctx):
                res = self.wrap_global_defined_check(res, receiver, ctx.module, node, ctx)
            elif ctx.current_function:
                ctx.current_function.call_deps.add((node, receiver, ctx.module))
        return res

    def translate_adt_decons(self, decons: PythonClass, node: ast.Attribute,
                             pos: Position, ctx: Context) -> StmtsAndExpr:
        """
        Deconstructs the ADT into the selected component, boxing the result
        if the projected component's type is ADT (recursive composition).
        """
        info = self.no_info(ctx)
        adt_type = self.viper.DomainType(decons.adt_domain_name, {}, [])

        # Translate receiver
        stmt, adt_obj = self.translate_expr(node.value, ctx)

        # Unbox ADT to be deconstructed
        unbox_func = self.viper.FuncApp(decons.fresh('unbox_' +
                                        decons.adt_domain_name),
                                        [adt_obj], pos, info, adt_type)

        # Calculate return type
        decons_type = (adt_type if decons.fields[node.attr].type == decons.adt_def
                       else self.translate_type(decons.fields[node.attr].type, ctx))

        # Translate deconstruction call
        decons_call = self.viper.DomainFuncApp(decons.fresh(decons.adt_prefix +
                                               decons.name + '_' + node.attr),
                                               [unbox_func], decons_type, pos,
                                               info, decons.adt_domain_name)

        # If returned type is ADT type, box it
        if decons_type == adt_type:
            decons_call = self.viper.FuncApp(decons.fresh('box_' +
                                             decons.adt_domain_name),
                                             [decons_call], pos, info,
                                             self.viper.Ref)

        return stmt, decons_call

    def translate_Attribute(self, node: ast.Attribute,
                            ctx: Context) -> StmtsAndExpr:
        position = self.to_position(node, ctx)
        target = self.get_target(node.value, ctx)
        func_name = get_func_name(node.value)
        if isinstance(target, PythonModule):
            target = self.get_target(node, ctx)
            if isinstance(target, PythonGlobalVar):
                return self.translate_global_var_reference(target, node, ctx)
            else:
                raise UnsupportedException(node)
        elif (isinstance(target, PythonClass) and
                      func_name != 'Result'):
            field = target.get_static_field(node.attr)
            field_func = self.translate_static_field_access(field, target,
                                                            node, ctx)
            return [], field_func
        elif (target is not None and hasattr(target, 'type') and
                  isinstance(target.type, UnionType) and
                  not isinstance(target.type, OptionalType)):
            stmt, receiver = self.translate_expr(node.value, ctx,
                                                 target_type=self.viper.Ref)
            guarded_field_access = []
            for recv_type in toposort_classes(target.type.get_types() - {None}):
                # Create a guard for accessing the field according to receiver's type
                field_guard = self.var_type_check(target.sil_name, recv_type, position, ctx)

                # Access the field of the specific type
                field = recv_type.get_field(node.attr)
                if isinstance(field, PythonField):
                    field_access = self.viper.FieldAccess(receiver,
                                                          field.actual_field.sil_field,
                                                          position, self.no_info(ctx))
                elif isinstance(field, PythonMethod):
                    assert False  # too lazy to implement now
                else:
                    field = recv_type.get_static_field(node.attr)
                    field_access = self.translate_static_field_access(field, receiver,
                                                                      node, ctx)

                guarded_field_access.append((field_guard, field_access))
            if len(guarded_field_access) == 1:
                _, field_access = guarded_field_access[0]
                return stmt, field_access
            else:
                return (stmt, chain_cond_exp(guarded_field_access, self.viper,
                                             position, self.no_info(ctx), ctx))
        else:
            # If the receiver is an ADT, attribute access is translated as deconstruction
            recv_type = self.get_type(node.value, ctx)
            if isinstance(recv_type, PythonClass) and recv_type.is_adt:
                return self.translate_adt_decons(recv_type, node, position, ctx)

            stmt, receiver = self.translate_expr(node.value, ctx,
                                                 target_type=self.viper.Ref)
            field = self._lookup_field(node, ctx)
            if isinstance(field, PythonGlobalVar):
                field_func = self.translate_static_field_access(field, receiver,
                                                                node, ctx)
                return [], field_func
            if isinstance(field, PythonMethod):
                # This is a reference to a property, so we translate it to a call of
                # the property getter function.
                target_type = self.translate_type(self.get_type(node.value, ctx), ctx)
                target_param = self.viper.LocalVarDecl('self', target_type, position,
                                                       self.no_info(ctx))
                property_type = self.translate_type(field.type, ctx)
                return stmt, self.viper.FuncApp(field.sil_name, [receiver], position,
                                                self.no_info(ctx), property_type,
                                                [target_param])
            return (stmt, self.viper.FieldAccess(receiver, field.sil_field,
                                                 position, self.no_info(ctx)))

    def translate_UnaryOp(self, node: ast.UnaryOp,
                          ctx: Context) -> StmtsAndExpr:
        if isinstance(node.op, ast.Not):
            stmt, expr = self.translate_expr(node.operand, ctx,
                                             target_type=self.viper.Bool)
            return (stmt, self.viper.Not(expr, self.to_position(node, ctx),
                                         self.no_info(ctx)))
        stmt, expr = self.translate_expr(node.operand, ctx,
                                         target_type=self.viper.Int)
        if isinstance(node.op, ast.USub):
            return (stmt, self.viper.Minus(expr, self.to_position(node, ctx),
                                           self.no_info(ctx)))
        else:
            raise UnsupportedException(node)

    def translate_IfExp(self, node: ast.IfExp, ctx: Context,
                        impure=False) -> StmtsAndExpr:
        position = self.to_position(node, ctx)
        cond_stmt, cond = self.translate_expr(node.test, ctx,
                                              target_type=self.viper.Bool)
        then_stmt, then = self.translate_expr(node.body, ctx,
                                              target_type=self._target_type,
                                              impure=impure)
        else_stmt, else_ = self.translate_expr(node.orelse, ctx,
                                               target_type=self._target_type,
                                               impure=impure)
        if then_stmt or else_stmt:
            then_block = self.translate_block(then_stmt, position,
                                              self.no_info(ctx))
            else_block = self.translate_block(else_stmt, position,
                                              self.no_info(ctx))
            if_stmt = self.viper.If(cond, then_block, else_block, position,
                                    self.no_info(ctx))
            bodystmt = [if_stmt]
        else:
            bodystmt = []
        cond_exp = self.viper.CondExp(cond, then, else_,
                                      self.to_position(node, ctx),
                                      self.no_info(ctx))
        return cond_stmt + bodystmt, cond_exp

    def translate_BinOp(self, node: ast.BinOp, ctx: Context) -> StmtsAndExpr:
        left_stmt, left = self.translate_expr(node.left, ctx)
        right_stmt, right = self.translate_expr(node.right, ctx)
        stmt = left_stmt + right_stmt
        left_type = self.get_type(node.left, ctx)
        right_type = self.get_type(node.right, ctx)
        op_stmt, result = self.translate_operator(left, right, left_type,
                                                  right_type, node, ctx)
        return stmt + op_stmt, result

    def _is_primitive_operation(self, op: ast.operator, left_type: PythonType,
                                right_type: PythonType) -> bool:
        """
        Determines if a binary operation with the given operand types can be
        translated as a native silver binary operation. True iff both types
        are identical and primitives.
        """
        # This is disabled for the moment because using the functions is advantageous
        # when using operations as triggers
        return False
        if type(op) not in self._primitive_operations:
            return False
        left_type_boxed = left_type.python_class.try_box()
        right_type_boxed = right_type.python_class.try_box()
        return (right_type_boxed.name in BOXED_PRIMITIVES and
                right_type_boxed.name == left_type_boxed.name)

    def _translate_primitive_operation(self, left: Expr, right: Expr,
                                       op_type: PythonType, op: ast.operator,
                                       pos: Position, ctx: Context) -> Expr:
        """
        Translates the binary operation consisting of the given operator and
        the given operands to a primitive Viper BinOp.
        """
        op = self._primitive_operations[type(op)]
        if op_type.python_class.try_box().name == INT_TYPE:
            wrap = self.to_int
        else:
            wrap = self.to_bool
        result = op(wrap(left, ctx), wrap(right, ctx), pos, self.no_info(ctx))
        return result

    def translate_operator(self, left: Expr, right: Expr, left_type: PythonType,
                           right_type: PythonType, node: ast.AST,
                           ctx: Context) -> StmtsAndExpr:
        """
        Translates the invocation of the binary operator of 'node' on the
        given two arguments, either to a primitive Silver operation or to a
        function or method call.
        """
        position = self.to_position(node, ctx)
        stmt = []
        if self._is_primitive_operation(node.op, left_type, right_type):
            result = self._translate_primitive_operation(left, right, left_type,
                                                         node.op, position, ctx)
            return stmt, result
        func_name = OPERATOR_FUNCTIONS[type(node.op)]
        call_stmt, call = self.get_func_or_method_call(left_type, func_name,
                                                       [left, right],
                                                       [left_type, right_type],
                                                       node, ctx)
        return stmt + call_stmt, call

    def is_thread_method_definition(self, node: ast.Compare, ctx: Context) -> bool:
        """
        Checks if the given comparison is a definition of a thread's target method,
        i.e., getMethod(t) == some_func.
        """
        if len(node.ops) != 1 or len(node.comparators) != 1:
            return False
        if not isinstance(node.ops[0], (ast.Eq, ast.Is)):
            return False
        for arg in (node.left, node.comparators[0]):
            if (isinstance(arg, ast.Call) and isinstance(arg.func, ast.Name) and
                        arg.func.id == 'getMethod'):
                return True
        return False

    def is_type_equality(self, node: ast.Compare, ctx: Context) -> bool:
        """
        Checks if a comparison checks the equality of the type of an object with
        something else (e.g., ``type(e1) == e2``), since these comparisons need special
        treatment.
        """
        if len(node.ops) != 1 or len(node.comparators) != 1:
            return False
        if not isinstance(node.ops[0], (ast.Eq, ast.Is, ast.NotEq, ast.IsNot)):
            return False
        for arg in (node.left, node.comparators[0]):
            if (isinstance(arg, ast.Call) and isinstance(arg.func, ast.Name) and
                        arg.func.id == 'type'):
                for other in (node.left, node.comparators[0]):
                    if other is not arg:
                        other_target = self.get_target(other, ctx)
                        if isinstance(other_target, PythonType):
                            return True
        return False

    def translate_type_equality(self, node: ast.Compare, ctx: Context) -> StmtsAndExpr:
        if (isinstance(node.left, ast.Call) and isinstance(node.left.func, ast.Name) and
                    node.left.func.id == 'type'):
            type_call = node.left
            type_literal = node.comparators[0]
        else:
            type_call = node.comparators[0]
            type_literal = node.left
        target = self.get_target(type_literal, ctx)
        assert isinstance(target, PythonType)
        call_stmt, call = self.translate_expr(type_call, ctx)
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        type_literal = self.type_factory.translate_type_literal(target.python_class,
                                                                pos, ctx, alias = call)
        if isinstance(node.ops[0], (ast.Is, ast.Eq)):
            func = self.viper.EqCmp
        else:
            func = self.viper.NeCmp
        comp = func(type_literal, call, pos, info)
        return [], comp

    def translate_thread_method_definition(self, node: ast.Compare,
                                           ctx: Context) -> StmtsAndExpr:
        if (isinstance(node.left, ast.Call) and isinstance(node.left.func, ast.Name) and
                    node.left.func.id == 'getMethod'):
            get_call = node.left
            method_literal = node.comparators[0]
        else:
            get_call = node.comparators[0]
            method_literal = node.left
        target_method = self.get_target(method_literal, ctx)
        if not isinstance(target_method, PythonMethod):
            raise InvalidProgramException(node, 'invalid.get.method')
        thread_stmt, thread = self.translate_expr(get_call.args[0], ctx)
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        method_id_type = self.viper.DomainType(METHOD_ID_DOMAIN, {}, [])
        thread_method = self.viper.DomainFuncApp(GET_METHOD_FUNC, [thread],
                                                 method_id_type, pos, info, THREAD_DOMAIN)
        method_literal = self.viper.DomainFuncApp(target_method.threading_id, [],
                                                  method_id_type,
                                                  pos, info, METHOD_ID_DOMAIN)
        comparison = self.viper.EqCmp(thread_method, method_literal, pos, info)
        return thread_stmt, comparison

    def translate_Compare(self, node: ast.Compare,
                          ctx: Context) -> StmtsAndExpr:
        if self.is_io_existential_defining_equality(node, ctx):
            self.define_io_existential(node, ctx)
            return ([], self.viper.TrueLit(self.to_position(node, ctx),
                                           self.no_info(ctx)))
        if self.is_wait_level_comparison(node, ctx):
            return self.translate_wait_level_comparison(node, ctx)
        if self.is_thread_method_definition(node, ctx):
            return self.translate_thread_method_definition(node, ctx)
        if self.is_type_equality(node, ctx):
            return self.translate_type_equality(node, ctx)
        if len(node.ops) != 1 or len(node.comparators) != 1:
            raise UnsupportedException(node)
        left_stmt, left = self.translate_expr(node.left, ctx)
        left_type = self.get_type(node.left, ctx)
        right_stmt, right = self.translate_expr(node.comparators[0], ctx)
        right_type = self.get_type(node.comparators[0], ctx)
        stmts = left_stmt + right_stmt
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)

        if self._is_primitive_operation(node.ops[0], left_type, right_type):
            result = self._translate_primitive_operation(left, right, left_type,
                                                         node.ops[0], position,
                                                         ctx)
            return stmts, result

        if isinstance(node.ops[0], ast.Is):
            return (stmts, self.viper.EqCmp(left, right, position, info))
        elif isinstance(node.ops[0], ast.IsNot):
            return (stmts, self.viper.NeCmp(left, right, position, info))
        elif isinstance(node.ops[0], (ast.In, ast.NotIn)):
            contains_stmts, contains_expr = self._translate_contains(
                left, right, left_type, right_type, node, ctx)
            stmts.extend(contains_stmts)
            return stmts, contains_expr
        if isinstance(node.ops[0], ast.Eq):
            int_compare = self.viper.EqCmp
            compare_func = '__eq__'
        elif isinstance(node.ops[0], ast.NotEq):
            int_compare = self.viper.NeCmp
            compare_func = '__ne__'
        elif isinstance(node.ops[0], ast.Gt):
            int_compare = self.viper.GtCmp
            compare_func = '__gt__'
        elif isinstance(node.ops[0], ast.GtE):
            int_compare = self.viper.GeCmp
            compare_func = '__ge__'
        elif isinstance(node.ops[0], ast.Lt):
            int_compare = self.viper.LtCmp
            compare_func = '__lt__'
        elif isinstance(node.ops[0], ast.LtE):
            int_compare = self.viper.LeCmp
            compare_func = '__le__'
        else:
            raise UnsupportedException(node.ops[0])
        if left_type.get_function(compare_func):
            comparison = self.get_function_call(left_type, compare_func,
                                                [left, right],
                                                [left_type, right_type],
                                                node, ctx)
        elif compare_func == '__ne__' and left_type.get_function('__eq__'):
            # The default behavior if __ne__ is not explicitly defined
            # is to invert the result of __eq__.
            call = self.get_function_call(left_type, '__eq__',
                                          [left, right],
                                          [left_type, right_type],
                                          node, ctx)
            comparison = self.viper.Not(self.to_bool(call, ctx, node.left),
                                        position, info)
        else:
            raise InvalidProgramException(node, 'undefined.comparison')
        return stmts, comparison

    def _translate_contains(
            self, left: Expr, right: Expr, left_type: PythonType,
            right_type: PythonType, node: ast.AST,
            ctx: Context) -> StmtsAndExpr:
        args = [right, left]
        arg_types = [right_type, left_type]
        app_stmt, app = self.get_func_or_method_call(
            right_type, '__contains__', args, arg_types, node, ctx)
        if isinstance(node.ops[0], ast.NotIn):
            app = self.viper.Not(
                app, self.to_position(node, ctx), self.no_info(ctx))
        return app_stmt, app

    def translate_NameConstant(self, node: ast.NameConstant,
                               ctx: Context) -> StmtsAndExpr:
        if node.value is True:
            return ([], self.viper.TrueLit(self.to_position(node, ctx),
                                           self.no_info(ctx)))
        elif node.value is False:
            return ([], self.viper.FalseLit(self.to_position(node, ctx),
                                            self.no_info(ctx)))
        elif node.value is None:
            return ([],
                    self.viper.NullLit(self.to_position(node, ctx),
                                       self.no_info(ctx)))
        else:
            raise UnsupportedException(node)

    def translate_BoolOp(self, node: ast.BoolOp, ctx: Context,
                         impure=False) -> StmtsAndExpr:
        assert isinstance(node.op, ast.Or) or isinstance(node.op, ast.And)

        position = self.to_position(node, ctx)
        info = self.no_info(ctx)

        statements_parts = []
        expression_parts = []
        bool_parts = []
        types_parts = []
        all_pure = True
        for value in node.values:
            typ = self.get_type(value, ctx)
            old_target = self._target_type
            # Translate expression to its original type, but with boolean
            # subexpressions.
            self._target_type = self.viper.Bool
            statements_part, expression_part = self._translate_only(
                value, ctx, impure)
            self._target_type = old_target
            # Get a version that is converted to a boolean. Unless we have
            # something impure.
            if self._is_pure(expression_part):
                bool_expression = self.to_bool(expression_part, ctx, value)
            else:
                bool_expression = expression_part
                all_pure = False
            if self._is_expression and statements_part:
                raise InvalidProgramException(node, 'not_expression')
            statements_parts.append(statements_part)
            if all_pure:
                expression_parts.append(self.to_ref(expression_part, ctx))
            bool_parts.append(bool_expression)
            types_parts.append(typ)

        all_bool = all(typ and typ.name == 'bool' for typ in types_parts)

        if isinstance(node.op, ast.And):
            operator = (
                lambda left, right:
                self.viper.And(left, right, position, info))
        else:
            operator = (
                lambda left, right:
                self.viper.Or(left, right, position, info))

        joined_bool_parts = [
            join_expressions(operator, bool_parts[:i + 1])
            for i in range(len(bool_parts))
            ]
        # If this is not an assertion (i.e. all parts are pure and there are non-boolean operands)
        if all_pure and not all_bool:
            # Instead of using Viper's And and Or, create an expression like
            # bool(lhs) ? lhs : rhs (or the other way round for Or).
            if isinstance(node.op, ast.And):
                operator = (
                    lambda left, left_bool, right:
                    self.viper.CondExp(left_bool, right, left, position, info)
                )
            else:
                operator = (
                    lambda left, left_bool, right:
                    self.viper.CondExp(left_bool, left, right, position, info)
                )
            joined_expression_parts = [
                join_three_expressions(operator, expression_parts[:i + 1],
                                       bool_parts[:i + 1], expression_parts[i])
                for i in range(len(bool_parts))
                ]

        statements = statements_parts[0]
        for i, part in enumerate(statements_parts[1:]):

            cond = joined_bool_parts[i]
            if isinstance(node.op, ast.Or):
                cond = self.viper.Not(cond, position, info)

            if part:
                then_block = self.translate_block(part, position, info)
                else_block = self.translate_block([], position, info)
                if_stmt = self.viper.If(cond, then_block, else_block,
                                        position, info)
                statements.append(if_stmt)
        if all_pure and not all_bool:
            res_val = joined_expression_parts[-1]
        else:
            res_val = joined_bool_parts[-1]
        return statements, res_val

    def translate_pythonvar_decl(self, var: PythonVar,
                                 ctx: Context) -> 'silver.ast.LocalVarDecl':
        """
        Creates a variable declaration for the given PythonVar.
        To be called during the processing phase by the Analyzer.
        """
        return self.viper.LocalVarDecl(var.sil_name,
                                       self.translate_type(var.type, ctx),
                                       self.no_position(ctx), self.no_info(ctx))

    def translate_pythonvar_ref(self, var: PythonVar, node: ast.AST,
                                ctx: Context) -> Expr:
        """
        Creates a variable reference for the given PythonVar.
        To be called during the processing phase by the Analyzer.
        """
        return self.viper.LocalVar(var.sil_name,
                                   self.translate_type(var.type, ctx),
                                   self.to_position(node, ctx),
                                   self.no_info(ctx))



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Silver statements."""


from typing import List

from nagini_translation.lib.program_nodes import (
    PythonVar,
)
from nagini_translation.lib.silver_nodes.base import (
    Expression,
    Statement,
)
from nagini_translation.lib.typedefs import (
    Info,
    Position,
    Stmt,
)


class Inhale(Statement):
    """Inhale statement."""

    def __init__(self, value: Expression) -> None:
        self._value = value

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Stmt:
        value = self._value.translate(translator, ctx, position, info)
        return translator.viper.Inhale(value, position, info)


class Exhale(Statement):
    """Exhale statement."""

    def __init__(self, value: Expression) -> None:
        self._value = value

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Stmt:
        value = self._value.translate(translator, ctx, position, info)
        return translator.viper.Exhale(value, position, info)


class Assert(Statement):
    """Assert statement."""

    def __init__(self, value: Expression) -> None:
        self._value = value

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Stmt:
        value = self._value.translate(translator, ctx, position, info)
        return translator.viper.Assert(value, position, info)


class Assign(Statement):
    """Assign statement."""

    def __init__(self, var: PythonVar, value: Expression) -> None:
        self._var = var
        self._value = value

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Stmt:
        value = self._value.translate(translator, ctx, position, info)
        return translator.viper.LocalVarAssign(
            self._var.ref(), value, position, info)


class If(Statement):
    """If statement."""

    def __init__(
            self, condition: 'BoolExpression', thn: List[Statement],
            els: List[Statement]) -> None:
        self._condition = condition
        self._then = thn
        self._else = els

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Stmt:
        condition = self._condition.translate(translator, ctx, position, info)
        then_body = translator.translate_block(
            [statement.translate(translator, ctx, position, info)
             for statement in self._then],
            position, info)
        else_body = translator.translate_block(
            [statement.translate(translator, ctx, position, info)
             for statement in self._else],
            position, info)
        return translator.viper.If(
            condition, then_body, else_body, position, info)


__all__ = (
    'Inhale',
    'Exhale',
    'Assert',
    'Assign',
    'If',
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

from typing import List, Tuple


Expr = 'silver.ast.Exp'

Stmt = 'silver.ast.Stmt'

StmtsAndExpr = Tuple[List[Stmt], Expr]

Var = 'silver.ast.LocalVar'

VarDecl = 'silver.ast.LocalVarDecl'

VarAssign = 'silver.ast.LocalVarAssign'

Domain = 'silver.ast.Domain'

DomainAxiom = 'silver.ast.DomainAxiom'

DomainFunc = 'silver.ast.DomainFunc'

DomainFuncApp = 'silver.ast.DomainFuncApp'

DomainType = 'silver.ast.DomainType'

Predicate = 'silver.ast.Predicate'

Program = 'silver.ast.Program'

Field = 'silver.ast.Field'

Function = 'silver.ast.Function'

Method = 'silver.ast.Method'

FuncApp = 'silver.ast.FuncApp'

TypeVar = 'silver.ast.TypeVar'

Type = 'silver.ast.Type'

Position = 'silver.ast.Position'

Info = 'silver.ast.Info'

Node = 'silver.ast.Node'

Seqn = 'silver.ast.Seqn'



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
import copy

from nagini_contracts.contracts import CONTRACT_WRAPPER_FUNCS
from nagini_translation.lib.constants import (
    BOOL_TYPE,
    BUILTIN_PREDICATES,
    GET_ARG_FUNC,
    GET_OLD_FUNC,
    GLOBAL_VAR_FIELD,
    INT_TYPE,
    JOINABLE_FUNC,
    METHOD_ID_DOMAIN,
    PMSET_TYPE,
    PRIMITIVES,
    PSEQ_TYPE,
    PSET_TYPE,
    RANGE_TYPE,
    THREAD_DOMAIN,
    THREAD_POST_PRED,
    THREAD_START_PRED,
)
from nagini_translation.lib.program_nodes import (
    PythonField,
    PythonGlobalVar,
    PythonMethod,
    PythonModule,
    PythonType,
    UnionType,
    toposort_classes,
    chain_cond_exp,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Position,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.util import (
    construct_lambda_prefix,
    find_loop_for_previous,
    get_func_name,
    InvalidProgramException,
    OldExpressionTransformer,
    pprint,
    UnsupportedException,
)
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator
from typing import List, Tuple


class ContractTranslator(CommonTranslator):

    def translate_contract(self, node: ast.AST, ctx: Context) -> Expr:
        """
        Generic visitor function for translating contracts (i.e. calls to
        contract functions)
        """
        method = 'translate_contract_' + node.__class__.__name__
        visitor = getattr(self, method, self.translate_generic)
        return visitor(node, ctx)

    def translate_contract_Call(self, node: ast.Call, ctx: Context) -> Expr:
        if get_func_name(node) in CONTRACT_WRAPPER_FUNCS:
            stmt, res = self.translate_expr(node.args[0], ctx, self.viper.Bool, True)
            if stmt:
                raise InvalidProgramException(node, 'purity.violated')
            return res
        else:
            raise UnsupportedException(node)

    def translate_contract_Expr(self, node: ast.Expr, ctx: Context) -> Expr:
        if isinstance(node.value, ast.Call):
            return self.translate_contract(node.value, ctx)
        else:
            raise UnsupportedException(node)

    def translate_result(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Result() contract function to a result
        expression.
        """
        assert len(node.args) == 0
        type = ctx.actual_function.type
        if not ctx.actual_function.pure:
            if ctx.result_var is None:
                raise InvalidProgramException(node, 'invalid.result')
            return [], ctx.result_var.ref(node, ctx)
        else:
            return ([], self.viper.Result(self.translate_type(type, ctx),
                                          self.to_position(node, ctx),
                                          self.no_info(ctx)))

    def translate_raised_exception(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the RaisedException() contract function to a
        expression.
        """
        assert len(node.args) == 0
        assert not ctx.actual_function.pure
        return [], ctx.error_var.ref(node, ctx)

    def _get_perm(self, node: ast.Call, ctx: Context) -> Expr:
        """
        Returns the permission for a Acc() contract function.
        """
        # Only one argument means implicit full permission
        if len(node.args) == 1:
            perm = self.viper.FullPerm(self.to_position(node, ctx),
                                       self.no_info(ctx))
        elif len(node.args) == 2:
            perm = self.translate_perm(node.args[1], ctx)
        else:
            # More than two arguments are invalid
            raise InvalidProgramException(node, 'invalid.contract.call')

        return perm

    def translate_builtin_predicate(self, node: ast.Call, perm: Expr,
                                    args: List[Expr], ctx: Context) -> Expr:
        name = node.func.id
        seq_ref = self.viper.SeqType(self.viper.Ref)
        set_ref = self.viper.SetType(self.viper.Ref)
        if name == 'list_pred':
            # field list_acc : Seq[Ref]
            field = self.viper.Field('list_acc', seq_ref, self.no_position(ctx),
                                     self.no_info(ctx))
        elif name == 'set_pred':
            # field set_acc : Set[Ref]
            field = self.viper.Field('set_acc', set_ref, self.no_position(ctx),
                                     self.no_info(ctx))
        elif name == 'dict_pred':
            # field dict_acc : Set[Ref]
            field = self.viper.Field('dict_acc', set_ref, self.no_position(ctx),
                                     self.no_info(ctx))
        elif name == 'MayStart':
            return self.translate_may_start(node, args, perm, ctx)
        elif name == 'ThreadPost':
            return self.translate_thread_post(node, args, perm, ctx)
        else:
            raise UnsupportedException(node)
        field_acc = self.viper.FieldAccess(args[0], field,
                                           self.to_position(node, ctx),
                                           self.no_info(ctx))
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        if ctx.perm_factor:
            perm = self.viper.PermMul(perm, ctx.perm_factor, pos, info)
        pred = self.viper.FieldAccessPredicate(field_acc, perm, pos, info)
        return pred

    def translate_may_start(self, node: ast.Call, args: List[Expr], perm: Expr,
                            ctx: Context) -> Expr:
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        pred_access = self.viper.PredicateAccess(args, THREAD_START_PRED, pos, info)
        access_pred = self.viper.PredicateAccessPredicate(pred_access, perm, pos,
                                                          info)
        return access_pred

    def translate_thread_post(self, node: ast.Call, args: List[Expr], perm: Expr,
                              ctx: Context) -> Expr:
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        pred_access = self.viper.PredicateAccess(args, THREAD_POST_PRED, pos, info)
        access_pred = self.viper.PredicateAccessPredicate(pred_access, perm, pos,
                                                          info)
        func = self.viper.FuncApp(JOINABLE_FUNC, args, pos, info, self.viper.Bool)
        conjunction = self.viper.And(access_pred, func, pos, info)
        return conjunction

    def translate_unwrapped_builtin_predicate(self, node: ast.Call, ctx: Context) -> Expr:
        args = []
        stmt, arg = self.translate_expr(node.args[0], ctx)
        if stmt:
            raise InvalidProgramException(node, 'purity.violated')
        perm = self.viper.FullPerm(self.no_position(ctx), self.no_info(ctx))
        return self.translate_builtin_predicate(node, perm, [arg], ctx)

    def translate_acc_predicate(self, node: ast.Call, perm: Expr,
                                ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Acc() contract function with a predicate call
        inside to a predicate access.
        """
        assert isinstance(node.args[0], ast.Call)
        call = node.args[0]
        # The predicate inside is a function call in python.
        args = []
        arg_stmts = []
        for arg in call.args:
            arg_stmt, arg_expr = self.translate_expr(arg, ctx)
            arg_stmts = arg_stmts + arg_stmt
            args.append(arg_expr)
        # Get the predicate inside the Acc()
        if isinstance(call.func, ast.Name):
            if call.func.id in BUILTIN_PREDICATES:
                return arg_stmts, self.translate_builtin_predicate(call, perm,
                                                                   args, ctx)
            else:
                pred = self.get_target(call.func, ctx)
        elif isinstance(call.func, ast.Attribute):
            receiver = self.get_target(call.func.value, ctx)
            if isinstance(receiver, PythonModule):
                pred = receiver.predicates[call.func.attr]
            else:
                rec_stmt, receiver = self.translate_expr(call.func.value, ctx)
                assert not rec_stmt
                receiver_class = self.get_type(call.func.value, ctx)
                name = call.func.attr
                pred = receiver_class.get_predicate(name)
                args = [receiver] + args
        else:
            raise UnsupportedException(node)
        if not (isinstance(pred, PythonMethod) and pred.predicate):
            raise InvalidProgramException(node, 'invalid.acc')
        pred_name = pred.sil_name
        # If the predicate is part of a family, find the correct version.
        if pred.cls:
            family_root = pred.cls
            while (family_root.superclass and
                   family_root.superclass.get_predicate(name)):
                family_root = family_root.superclass
            pred_name = family_root.get_predicate(name).sil_name
        return arg_stmts, self.create_predicate_access(pred_name, args, perm,
                                                       node, ctx)

    def translate_acc_field(self, node: ast.Call, perm: Expr,
                            ctx: Context) -> StmtsAndExpr:
        assert isinstance(node.args[0], ast.Attribute)
        field = self.get_target(node.args[0], ctx)
        if not isinstance(field, PythonField):
            raise InvalidProgramException(node, 'invalid.acc')
        stmt, field_acc = self.translate_expr(node.args[0], ctx)
        if stmt:
            raise InvalidProgramException(node, 'purity.violated')
        field_type = self.get_type(node.args[0], ctx)
        pred = self._translate_acc_field(field_acc, field_type, perm,
                                         self.to_position(node, ctx), ctx)
        return [], pred

    def translate_acc_global(self, node: ast.Call, perm: Expr,
                            ctx: Context) -> StmtsAndExpr:
        """
        Translates an access permission to a global variable.
        """
        var = self.get_target(node.args[0], ctx)
        if not isinstance(var, PythonGlobalVar):
            raise InvalidProgramException(node, 'invalid.acc')
        if var.is_final:
            raise InvalidProgramException(node, 'permission.to.final.var')
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        var_func = self.viper.FuncApp(var.sil_name, [], pos, info, self.viper.Ref, [])
        var_type = self.translate_type(var.type, ctx)
        field = self.viper.Field(GLOBAL_VAR_FIELD, var_type, pos, info)
        field_acc = self.viper.FieldAccess(var_func, field, pos, info)
        pred = self.viper.FieldAccessPredicate(field_acc, perm, pos, info)

        # Add type information
        if var.type.name not in PRIMITIVES:
            type_info = self.type_check(field_acc, var.type,
                                        self.no_position(ctx), ctx)
            pred = self.viper.And(pred, type_info, pos, info)
        return [], pred

    def _translate_acc_field(self, field_acc: Expr, field_type: PythonType,
                             perm: Expr, pos: Position, ctx: Context) -> StmtsAndExpr:
        info = self.no_info(ctx)
        if ctx.perm_factor:
            perm = self.viper.PermMul(perm, ctx.perm_factor, pos, info)
        pred = self.viper.FieldAccessPredicate(field_acc, perm,
                                               pos, info)
        # Add type information
        if field_type.name not in PRIMITIVES:
            type_info = self.type_check(field_acc, field_type,
                                        self.no_position(ctx), ctx)
            pred = self.viper.And(pred, type_info, pos, info)
        return pred

    def translate_may_set(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to MaySet().
        """
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        stmt, rec = self.translate_expr(node.args[0], ctx)
        rec_type = self.get_type(node.args[0], ctx)
        if stmt:
            raise InvalidProgramException(node.args[0], 'purity.violated')
        if not isinstance(node.args[1], ast.Str):
            raise InvalidProgramException(node.args[1], 'invalid.may.set')
        field = rec_type.get_field(node.args[1].s)
        if not field:
            raise InvalidProgramException(node.args[1], 'invalid.may.set')
        may_set_pred = self.get_may_set_predicate(rec, field, ctx, pos)
        sil_field = self.viper.Field(field.sil_name, self.translate_type(field.type, ctx),
                                     pos, info)
        field_acc = self.viper.FieldAccess(rec, sil_field, pos, info)
        full_perm = self.viper.FullPerm(pos, info)
        normal_acc = self._translate_acc_field(field_acc, field.type, full_perm, pos, ctx)
        normal_perm = self.viper.CurrentPerm(field_acc, pos, info)
        have_normal_perm = self.viper.PermGtCmp(normal_perm, self.viper.NoPerm(pos, info),
                                                pos, info)
        result_ex = self.viper.CondExp(have_normal_perm, normal_acc, may_set_pred, pos,
                                       info)
        unknown = self.get_unknown_bool(ctx)
        result_in = self.viper.CondExp(unknown, normal_acc, may_set_pred, pos, info)
        return [], self.viper.InhaleExhaleExp(result_in, result_ex, pos, info)

    def translate_may_create(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to MayCreate().
        """
        pos = self.to_position(node, ctx)
        stmt, rec = self.translate_expr(node.args[0], ctx)
        rec_type = self.get_type(node.args[0], ctx)
        if stmt:
            raise InvalidProgramException(node.args[0], 'purity.violated')
        if not isinstance(node.args[1], ast.Str):
            raise InvalidProgramException(node.args[1], 'invalid.may.create')
        field = rec_type.get_field(node.args[1].s)
        if not field:
            raise InvalidProgramException(node.args[1], 'invalid.may.create')
        return [], self.get_may_set_predicate(rec, field, ctx, pos)

    def translate_assert(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to Assert().
        """
        assert len(node.args) == 1
        stmt, expr = self.translate_expr(node.args[0], ctx, self.viper.Bool, True)
        assertion = self.viper.Assert(expr, self.to_position(node, ctx),
                                      self.no_info(ctx))
        return stmt + [assertion], None

    def translate_assume(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to Assume().
        """
        assert len(node.args) == 1
        stmt, expr = self.translate_expr(node.args[0], ctx, self.viper.Bool, False)
        assertion = self.viper.Inhale(expr, self.to_position(node, ctx),
                                      self.no_info(ctx))
        return stmt + [assertion], None

    def translate_implies(self, node: ast.Call, ctx: Context,
                          impure=False) -> StmtsAndExpr:
        """
        Translates a call to the Implies() contract function.
        """
        if len(node.args) != 2:
            raise InvalidProgramException(node, 'invalid.contract.call')
        cond_stmt, cond = self.translate_expr(node.args[0], ctx,
                                              target_type=self.viper.Bool)
        then_stmt, then = self.translate_expr(node.args[1], ctx,
                                              target_type=self.viper.Bool, impure=impure)
        implication = self.viper.Implies(cond, then,
                                         self.to_position(node, ctx),
                                         self.no_info(ctx))
        return (cond_stmt + then_stmt, implication)

    def translate_old(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Old() contract function.
        """
        if len(node.args) != 1:
            raise InvalidProgramException(node, 'invalid.contract.call')

        if ctx.old_expr_aliases:
            normalizer = OldExpressionTransformer()
            normalizer.arg_names = [a for a in ctx.actual_function._args]
            normalized = normalizer.visit(copy.deepcopy(node.args[0]))
            key = pprint(normalized)
            if key in ctx.old_expr_aliases:
                return [], ctx.old_expr_aliases[key]

        stmt, exp = self.translate_expr(node.args[0], ctx)
        res = self.viper.Old(exp, self.to_position(node, ctx),
                             self.no_info(ctx))
        return stmt, res

    def translate_fold(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Fold() contract function.
        """
        if len(node.args) != 1:
            raise InvalidProgramException(node, 'invalid.contract.call')
        if not isinstance(node.args[0], ast.Call):
            raise InvalidProgramException(node, 'invalid.contract.call')
        if get_func_name(node.args[0]) in ('Acc', 'Rd'):
            pred_call = node.args[0].args[0]
        else:
            pred_call = node.args[0]
        target_pred = self.get_target(pred_call, ctx)
        if (target_pred and
                (not isinstance(target_pred, PythonMethod) or not target_pred.predicate)):
            raise InvalidProgramException(node, 'invalid.contract.call')
        pred_stmt, pred = self.translate_expr(node.args[0], ctx,
                                              self.viper.Bool, True)
        if self._is_family_fold(node):
            # Predicate called on receiver, so it belongs to a family
            if ctx.ignore_family_folds:
                return [], None
        if pred_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        fold = self.viper.Fold(pred, self.to_position(node, ctx),
                               self.no_info(ctx))
        return [fold], None

    def _is_family_fold(self, fold: ast.AST) -> bool:
        if isinstance(fold.args[0], ast.Call) and \
                isinstance(fold.args[0].func, ast.Attribute):
            return True
        if (isinstance(fold.args[0], ast.Call) and
                isinstance(fold.args[0].func, ast.Name) and
                fold.args[0].func.id == 'Acc' and
                isinstance(fold.args[0].args[0], ast.Attribute)):
            return True
        return False

    def translate_previous(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        arg = node.args[0]
        if not isinstance(arg, ast.Name):
            raise InvalidProgramException(node, 'invalid.previous')
        loop = find_loop_for_previous(node, arg.id)
        if not loop:
            raise InvalidProgramException(node, 'invalid.previous')
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        iterator = ctx.loop_iterators[loop].ref()
        list_field = self.viper.Field('__previous', self.viper.SeqType(self.viper.Ref),
                                      pos, info)
        field_acc = self.viper.FieldAccess(iterator, list_field, pos, info)
        seq_type = ctx.module.global_module.classes[PSEQ_TYPE]
        content_type = self.get_type(node.args[0], ctx)
        type_lit = self.type_factory.translate_type_literal(content_type, pos, ctx)
        res = self.get_function_call(seq_type, '__create__', [field_acc, type_lit],
                                     [None, None], node, ctx)
        return [], res

    def translate_unfold(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Unfold() contract function.
        """
        if len(node.args) != 1:
            raise InvalidProgramException(node, 'invalid.contract.call')
        if not isinstance(node.args[0], ast.Call):
            raise InvalidProgramException(node, 'invalid.contract.call')
        if get_func_name(node.args[0]) in ('Acc', 'Rd'):
            pred_call = node.args[0].args[0]
        else:
            pred_call = node.args[0]
        target_pred = self.get_target(pred_call, ctx)
        if (target_pred and
                (not isinstance(target_pred, PythonMethod) or not target_pred.predicate)):
            raise InvalidProgramException(node, 'invalid.contract.call')
        pred_stmt, pred = self.translate_expr(node.args[0], ctx,
                                              self.viper.Bool, True)
        if self._is_family_fold(node):
            # Predicate called on receiver, so it belongs to a family
            if ctx.ignore_family_folds:
                return [], None
        if pred_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        unfold = self.viper.Unfold(pred, self.to_position(node, ctx),
                                   self.no_info(ctx))
        return [unfold], None

    def translate_unfolding(self, node: ast.Call, ctx: Context,
                            impure=False) -> StmtsAndExpr:
        """
        Translates a call to the Unfolding() contract function.
        """
        if len(node.args) != 2:
            raise InvalidProgramException(node, 'invalid.contract.call')
        if not isinstance(node.args[0], ast.Call):
            raise InvalidProgramException(node, 'invalid.contract.call')
        if get_func_name(node.args[0]) in ('Acc', 'Rd'):
            pred_call = node.args[0].args[0]
        else:
            pred_call = node.args[0]
        target_pred = self.get_target(pred_call, ctx)
        if (target_pred and
                (not isinstance(target_pred, PythonMethod) or not target_pred.predicate)):
            raise InvalidProgramException(node, 'invalid.contract.call')

        pred_stmt, pred = self.translate_expr(node.args[0], ctx,
                                              self.viper.Bool, True)
        if pred_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        expr_stmt, expr = self.translate_expr(node.args[1], ctx)
        if expr_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        expr = self.unwrap(expr)
        unfold = self.viper.Unfolding(pred, expr, self.to_position(node, ctx),
                                      self.no_info(ctx))
        return expr_stmt, unfold

    def translate_low(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Low() contract function.
        """
        return [], self.viper.TrueLit(self.to_position(node, ctx),
                                      self.no_info(ctx))

    def translate_lowval(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the LowVal() contract function.
        """
        return self.translate_low(node, ctx)

    def translate_lowevent(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the LowEvent() contract function.
        """
        return [], self.viper.TrueLit(self.to_position(node, ctx),
                                      self.no_info(ctx))

    def translate_declassify(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Declassify() contract function.
        """
        return [self.translate_block([], self.no_position(ctx), self.no_info(ctx))], None

    def translate_terminates_sif(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the TerminatesSif() contract function.
        """
        cond_stmts, cond = self.translate_expr(node.args[0], ctx)
        rank_stmts, rank = self.translate_expr(node.args[1], ctx)
        if cond_stmts or rank_stmts:
            raise InvalidProgramException(node, 'purity.violated')
        return self.translator.obligation_translator._translate_must_terminate(node, ctx)

    def _translate_triggers(self, body: ast.AST, node: ast.Call,
                            ctx: Context) -> List['silver.ast.Trigger']:
        """
        Assuming the given body is a tuple whose second element is a list
        literal containing any number of list literals containing expressions,
        translates those to a list of triggers.
        """
        if (not isinstance(body, ast.Tuple) or
                    len(body.elts) != 2):
            raise InvalidProgramException(node, 'invalid.trigger')
        trigger_node = body.elts[1]
        triggers = []
        if not isinstance(trigger_node, ast.List):
            raise InvalidProgramException(node, 'invalid.trigger')
        outer = trigger_node

        if outer.elts:
            for el in outer.elts:
                trigger = []
                if not isinstance(el, ast.List):
                    raise InvalidProgramException(el, 'invalid.trigger')
                for inner in el.elts:
                    if (isinstance(inner, ast.Compare) and len(inner.ops) == 1 and
                            isinstance(inner.ops[0], ast.In)):
                        # Use the less complex and more efficient trigger translation we
                        # also use for the domain of the forall quantifier.
                        assert len(inner.comparators) == 1
                        lhs_stmt, lhs = self.translate_expr(inner.left, ctx)
                        part_stmt, part, valid = self._create_quantifier_contains_expr(
                            lhs, inner.comparators[0], ctx)
                        if part_stmt:
                            raise InvalidProgramException(inner,
                                                          'purity.violated')
                        if valid and not part_stmt and not lhs_stmt:
                            trigger.append(part)
                            continue

                    part_stmt, part = self.translate_expr(inner, ctx)
                    if part_stmt:
                        raise InvalidProgramException(inner,
                                                      'purity.violated')
                    part = self.unwrap(part)
                    trigger.append(part)
                trigger = self.viper.Trigger(trigger, self.no_position(ctx),
                                             self.no_info(ctx))
                triggers.append(trigger)
        return triggers

    def _create_quantifier_contains_expr(self, e: Expr,
                                         domain_node: ast.AST,
                                         ctx: Context,
                                         trigger=False) -> Tuple[List[Stmt], Expr, bool]:
        """
        Creates the left hand side of the implication in a quantifier
        expression, which says that e is an element of the given domain.
        The last return value specifies if the returned expression is
        recommended to be used as a trigger.
        """
        domain_old = False
        if (isinstance(domain_node, ast.Call) and
                    get_func_name(domain_node) == 'Old'):
            domain_old = True
            domain_node = domain_node.args[0]
        ref_var = self.to_ref(e, ctx)
        pos = self.to_position(domain_node, ctx)
        info = self.no_info(ctx)

        dom_target = self.get_target(domain_node, ctx)

        if isinstance(dom_target, PythonType):
            result = self.type_check(ref_var, dom_target, pos, ctx, False)
            # Not recommended as a trigger, since it's very broad and will get triggered
            # a lot.
            return [], result, False
        dom_stmt, domain = self.translate_expr(domain_node, ctx)
        dom_type = self.get_type(domain_node, ctx)
        result = self.get_quantifier_lhs(ref_var, dom_type, domain, domain_node, ctx, pos,
                                         trigger)
        if domain_old:
            result = self.viper.Old(result, pos, info)
        return dom_stmt, result, True

    def translate_to_sequence(self, node: ast.Call,
                              ctx: Context) -> StmtsAndExpr:
        coll_type = self.get_type(node.args[0], ctx)
        stmt, arg = self.translate_expr(node.args[0], ctx)
        # Use the same sequence conversion as for iterating over the
        # iterable (which gives no information about order for unordered types).
        seq_call = self.get_sequence(coll_type, arg, None, node, ctx)
        seq_class = ctx.module.global_module.classes[PSEQ_TYPE]
        if coll_type.name == RANGE_TYPE:
            type_arg = ctx.module.global_module.classes[INT_TYPE]
        else:
            type_arg = coll_type.type_args[0]
        position = self.to_position(node, ctx)
        type_lit = self.type_factory.translate_type_literal(type_arg, position,
                                                            ctx)
        result = self.get_function_call(seq_class, '__create__',
                                        [seq_call, type_lit], [None, None],
                                        node, ctx)
        return stmt, result

    def translate_sequence(self, node: ast.Call,
                           ctx: Context) -> StmtsAndExpr:
        seq_type = self.get_type(node, ctx)
        viper_type = self.translate_type(seq_type.type_args[0], ctx)
        val_stmts = []
        if node.args:
            vals = []
            for arg in node.args:
                arg_stmt, arg_val = self.translate_expr(arg, ctx,
                    target_type=viper_type)
                val_stmts += arg_stmt
                vals.append(arg_val)
            result = self.viper.ExplicitSeq(vals, self.to_position(node,
                                                                   ctx),
                                            self.no_info(ctx))
        else:
            result = self.viper.EmptySeq(viper_type,
                                         self.to_position(node, ctx),
                                         self.no_info(ctx))
        type_arg = seq_type.type_args[0]
        position = self.to_position(node, ctx)
        type_lit = self.type_factory.translate_type_literal(type_arg, position,
                                                            ctx)
        result = self.get_function_call(seq_type.cls, '__create__',
                                        [result, type_lit], [None, None], node,
                                        ctx)
        return val_stmts, result

    def translate_pset(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        set_type = self.get_type(node, ctx)

        viper_type = self.translate_type(set_type.type_args[0], ctx)
        val_stmts = []

        if node.args:
            vals = []
            for arg in node.args:
                arg_stmt, arg_val = self.translate_expr(arg, ctx,
                    target_type = viper_type)
                val_stmts += arg_stmt
                vals.append(arg_val)
            result = self.viper.ExplicitSet(vals, self.to_position(node, ctx),
                                            self.no_info(ctx))
        else:
            result = self.viper.EmptySet(viper_type,
                                         self.to_position(node, ctx),
                                         self.no_info(ctx))
        type_arg = set_type.type_args[0]
        position = self.to_position(node, ctx)
        type_lit = self.type_factory.translate_type_literal(type_arg, position,
                                                            ctx)
        result = self.get_function_call(set_type.cls, '__create__',
                                        [result, type_lit], [None, None], node,
                                        ctx)
        return val_stmts, result

    def translate_mset(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        mset_type = self.get_type(node, ctx)

        viper_type = self.translate_type(mset_type.type_args[0], ctx)
        val_stmts = []

        if node.args:
            vals = []
            for arg in node.args:
                arg_stmt, arg_val = self.translate_expr(arg, ctx,
                    target_type = viper_type)
                val_stmts += arg_stmt
                vals.append(arg_val)
            result = self.viper.ExplicitMultiset(vals, self.to_position(node, ctx),
                                                 self.no_info(ctx))
        else:
            result = self.viper.EmptyMultiset(viper_type,
                                         self.to_position(node, ctx),
                                         self.no_info(ctx))
        type_arg = mset_type.type_args[0]
        position = self.to_position(node, ctx)
        type_lit = self.type_factory.translate_type_literal(type_arg, position,
                                                            ctx)
        result = self.get_function_call(mset_type.cls, '__create__',
                                        [result, type_lit], [None, None], node,
                                        ctx)
        return val_stmts, result

    def translate_get_arg(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        thread_stmt, thread = self.translate_expr(node.args[0], ctx)
        index_stmt, index = self.translate_expr(node.args[1], ctx,
                                                target_type = self.viper.Int)
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        method_id_type = self.viper.DomainType(METHOD_ID_DOMAIN, {}, [])
        func = self.viper.DomainFuncApp(GET_ARG_FUNC, [thread, index], self.viper.Ref,
                                        pos, info, THREAD_DOMAIN)
        return thread_stmt + index_stmt, func

    def translate_get_old(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        stmt, thread = self.translate_expr(node.args[0], ctx)
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        old_string = pprint(node.args[1])
        index = self._get_string_value(old_string)
        index = self.viper.IntLit(index, pos, info)
        func = self.viper.DomainFuncApp(GET_OLD_FUNC, [thread, index], self.viper.Ref,
                                        pos, info, THREAD_DOMAIN)
        return stmt, func

    def translate_joinable(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        stmt, thread = self.translate_expr(node.args[0], ctx)
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        func = self.viper.FuncApp(JOINABLE_FUNC, [thread], pos, info, self.viper.Bool)
        return stmt, func

    def translate_let(self, node: ast.Call, ctx: Context,
                      impure : bool = False) -> StmtsAndExpr:
        type = self.get_target(node.args[1], ctx)
        if not isinstance(type, PythonType) or not isinstance(node.args[2], ast.Lambda):
            raise InvalidProgramException(node, 'invalid.let')
        lambda_ = node.args[2]
        lambda_prefix = construct_lambda_prefix(lambda_.lineno,
                                                getattr(lambda_, 'col_offset',
                                                        None))
        lambda_prefix += '$'
        arg = lambda_.args.args[0]
        var = ctx.actual_function.get_variable(lambda_prefix + arg.arg)

        exp_stmt, exp_val = self.translate_expr(node.args[0], ctx)

        ctx.set_alias(arg.arg, var, None)

        body_stmt, body_val = self.translate_expr(lambda_.body, ctx, impure=impure)

        ctx.remove_alias(arg.arg)
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        let = self.viper.Let(var.decl, exp_val, body_val, pos, info)
        return exp_stmt + body_stmt, let

    def translate_forall(self, node: ast.Call, ctx: Context,
                         impure=False) -> StmtsAndExpr:
        domain_node = node.args[0]

        lambda_ = node.args[1]
        variables = []
        lambda_prefix = construct_lambda_prefix(lambda_.lineno,
                                                getattr(lambda_, 'col_offset',
                                                        None))
        lambda_prefix += '$'
        arg = lambda_.args.args[0]
        var = ctx.actual_function.get_variable(lambda_prefix + arg.arg)
        variables.append(var.decl)

        ctx.set_alias(arg.arg, var, None)
        if isinstance(lambda_.body, ast.Tuple):
            if not len(lambda_.body.elts) == 2:
                raise InvalidProgramException(node, 'invalid.forall')
            body_stmt, rhs = self.translate_expr(lambda_.body.elts[0], ctx,
                                                 self.viper.Bool, impure)

            triggers = self._translate_triggers(lambda_.body, node, ctx)
        else:
            body_type = self.get_type(lambda_.body, ctx)
            if not body_type or body_type.name != BOOL_TYPE:
                raise InvalidProgramException(node, 'invalid.forall')
            body_stmt, rhs = self.translate_expr(lambda_.body, ctx,
                                                 self.viper.Bool, impure)
            triggers = []

        ctx.remove_alias(arg.arg)
        if body_stmt:
            raise InvalidProgramException(node, 'purity.violated')

        dom_stmt, lhs, always_use = self._create_quantifier_contains_expr(var.ref(),
                                                                          domain_node,
                                                                          ctx)
        if dom_stmt:
            raise InvalidProgramException(domain_node,
                                          'purity.violated')
        lhs = self.unwrap(lhs)

        implication = self.viper.Implies(lhs, rhs, self.to_position(node, ctx),
                                         self.no_info(ctx))
        if always_use or not triggers:
            # Add lhs of the implication, which the user cannot write directly
            # in this exact form.
            # If we always do this, we apparently deactivate the automatically
            # generated triggers and things are actually worse.
            # Change: We always do this now.
            try:
                # Depending on the collection expression, this doesn't always
                # work (malformed trigger); in that case, we just don't do it.
                lhs_trigger = self.viper.Trigger([lhs], self.no_position(ctx),
                                                 self.no_info(ctx))
                triggers = [lhs_trigger] + triggers
            except Exception:
                pass
        var_type_check = self.type_check(var.ref(), var.type,
                                         self.no_position(ctx), ctx, False)
        implication = self.viper.Implies(var_type_check, implication,
                                         self.to_position(node, ctx),
                                         self.no_info(ctx))
        forall = self.viper.Forall(variables, triggers, implication,
                                   self.to_position(node, ctx),
                                   self.no_info(ctx))
        return dom_stmt, forall

    def translate_contractfunc_call(self, node: ast.Call, ctx: Context,
                                    impure=False, statement=False) -> StmtsAndExpr:
        """
        Translates calls to contract functions like Result() and Acc()
        """
        func_name = get_func_name(node)
        if func_name == 'Result':
            return self.translate_result(node, ctx)
        elif func_name == 'RaisedException':
            return self.translate_raised_exception(node, ctx)
        elif func_name in ('Acc', 'Rd', 'Wildcard'):
            if not impure:
                raise InvalidProgramException(node, 'invalid.contract.position')
            if func_name == 'Rd':
                perm = self.get_arp_for_context(node, ctx)
            elif func_name == 'Wildcard':
                perm = self.viper.WildcardPerm(self.to_position(node, ctx),
                                               self.no_info(ctx))
            else:
                perm = self._get_perm(node, ctx)
            if isinstance(node.args[0], ast.Call):
                return self.translate_acc_predicate(node, perm, ctx)
            else:
                if isinstance(node.args[0], ast.Attribute):
                    type = self.get_type(node.args[0].value, ctx)
                    if isinstance(type, UnionType):
                        guarded_field_access = []
                        stmt, receiver = self.translate_expr(node.args[0].value, ctx)
                        for recv_type in toposort_classes(type.get_types() - {None}):
                            target = self.get_target(node.args[0].value, ctx)
                            field_guard = self.var_type_check(target.sil_name,
                                                              recv_type,
                                                              self.to_position(node, ctx),
                                                              ctx)
                            field = recv_type.get_field(node.args[0].attr).actual_field
                            field_access = self.viper.FieldAccess(receiver, field.sil_field,
                                                                  self.to_position(node, ctx),
                                                                  self.no_info(ctx))
                            field_acc = self._translate_acc_field(field_access,
                                                                  field.type, perm,
                                                                  self.to_position(node, ctx),
                                                                  ctx)
                            guarded_field_access.append((field_guard, field_acc))
                        if len(guarded_field_access) == 1:
                            _, field_acc = guarded_field_access[0]
                            return stmt, field_acc
                        else:
                            return (stmt, chain_cond_exp(guarded_field_access, self.viper,
                                                         self.to_position(node, ctx),
                                                         self.no_info(ctx), ctx))
                target = self.get_target(node.args[0], ctx)
                if isinstance(target, PythonField):
                    return self.translate_acc_field(node, perm, ctx)
                else:
                    if not isinstance(target, PythonGlobalVar):
                        raise InvalidProgramException(node, 'invalid.acc')
                    return self.translate_acc_global(node, perm, ctx)
        elif func_name in BUILTIN_PREDICATES:
            return [], self.translate_unwrapped_builtin_predicate(node, ctx)
        elif func_name == 'MaySet':
            return self.translate_may_set(node, ctx)
        elif func_name == 'MayCreate':
            return self.translate_may_create(node, ctx)
        elif func_name in ('Assert', 'Assume', 'Fold', 'Unfold'):
            if not statement:
                raise InvalidProgramException(node, 'invalid.contract.position')
            if func_name == 'Assert':
                return self.translate_assert(node, ctx)
            elif func_name == 'Assume':
                return self.translate_assume(node, ctx)
            elif func_name == 'Fold':
                return self.translate_fold(node, ctx)
            elif func_name == 'Unfold':
                return self.translate_unfold(node, ctx)
        elif func_name == 'Implies':
            return self.translate_implies(node, ctx, impure)
        elif func_name == 'Old':
            return self.translate_old(node, ctx)
        elif func_name == 'Unfolding':
            return self.translate_unfolding(node, ctx, impure)
        elif func_name == 'Low':
            return self.translate_low(node, ctx)
        elif func_name == 'LowVal':
            return self.translate_lowval(node, ctx)
        elif func_name == 'LowEvent':
            return self.translate_lowevent(node, ctx)
        elif func_name == 'Declassify':
            return self.translate_declassify(node, ctx)
        elif func_name == 'TerminatesSif':
            return self.translate_terminates_sif(node, ctx)
        elif func_name == 'Forall':
            return self.translate_forall(node, ctx, impure)
        elif func_name == 'Previous':
            return self.translate_previous(node, ctx)
        elif func_name == 'Let':
            return self.translate_let(node, ctx, impure)
        elif func_name == PSEQ_TYPE:
            return self.translate_sequence(node, ctx)
        elif func_name == PSET_TYPE:
            return self.translate_pset(node, ctx)
        elif func_name == PMSET_TYPE:
            return self.translate_mset(node, ctx)
        elif func_name == 'ToSeq':
            return self.translate_to_sequence(node, ctx)
        elif func_name == 'Joinable':
            return self.translate_joinable(node, ctx)
        elif func_name == 'getArg':
            return self.translate_get_arg(node, ctx)
        elif func_name == 'getOld':
            return self.translate_get_old(node, ctx)
        elif func_name == 'getMethod':
            raise InvalidProgramException(node, 'invalid.get.method.use')
        elif func_name == 'arg':
            raise InvalidProgramException(node, 'invalid.arg.use')
        else:
            raise UnsupportedException(node)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_contracts.contracts import CONTRACT_FUNCS
from nagini_translation.lib.constants import (
    BOOL_TYPE,
    BUILTINS,
    BYTES_TYPE,
    DICT_TYPE,
    INT_TYPE,
    LIST_TYPE,
    OBJECT_TYPE,
    OPERATOR_FUNCTIONS,
    PMSET_TYPE,
    PSEQ_TYPE,
    PSET_TYPE,
    RANGE_TYPE,
    SET_TYPE,
    STRING_TYPE,
    TUPLE_TYPE,
)
from nagini_translation.lib.program_nodes import (
    ContainerInterface,
    GenericType,
    OptionalType,
    PythonClass,
    PythonField,
    PythonIOOperation,
    PythonMethod,
    PythonModule,
    PythonNode,
    PythonType,
    PythonVarBase,
    TypeVar,
    UnionType,
)
from nagini_translation.lib.util import (
    get_func_name,
    InvalidProgramException,
    UnsupportedException,
)
from typing import List, Optional


def get_target(node: ast.AST,
               containers: List[ContainerInterface],
               container: PythonNode, type: bool = False) -> Optional[PythonNode]:
    """
    Finds the PythonNode that the given ``node`` refers to, e.g. a PythonClass
    or a PythonVar, if the immediate container (e.g. a PythonMethod) of the node
    is ``container``, by looking in the given ``containers`` (can be e.g.
    PythonMethods, the Context, PythonModules, etc).
    If the ``type`` parameter is set, will also consider string literals as potential
    references.
    """
    if isinstance(node, ast.Name):
        return find_entry(node.id, True, containers)
    elif type and isinstance(node, ast.Str):
        return find_entry(node.s, True, containers)
    elif isinstance(node, ast.Call):
        # For calls, we return the type of the result of the call
        func_name = get_func_name(node)
        if (container and func_name == 'Result' and
                isinstance(container, PythonMethod)):
            # In this case the immediate container must be a method, and we
            # return its result type
            return container.type
        elif (container and func_name == 'super' and
                  isinstance(container, PythonMethod)):
            # Return the type of the current method's superclass
            return container.cls.superclass
        elif func_name == 'cast':
            return None
        return get_target(node.func, containers, container)
    elif isinstance(node, ast.Attribute):
        # Find the type of the LHS, so that we can look through its members.
        lhs = get_type(node.value, containers, container)
        if isinstance(lhs, OptionalType):
            lhs = lhs.optional_type
        if isinstance(lhs, UnionType):
            # When receiver's type is union, a method call have multiple
            # targets, therefore None is returned in such cases
            return None
        if isinstance(lhs, GenericType) and lhs.name == 'type':
            # For direct references to type objects, we want to lookup things
            # defined in the class. So instead of type[C], we want to look in
            # class C directly here.
            lhs = lhs.type_args[0]
        if isinstance(lhs, GenericType):
            # Use the class, since we want to look for members.
            lhs = lhs.cls
        # Now collect all containers we have to look through
        containers = []
        if isinstance(lhs, PythonModule):
            # We have to look through all included modules as well, but not
            # through the global one, since it makes no sense to refer to
            # global stuff by looking in a different module
            containers.extend(lhs.get_included_modules(include_global=False))
        else:
            containers.append(lhs)
        while (isinstance(containers[-1], PythonClass) and
                   containers[-1].superclass):
            # If we're looking in a class, add all superclasses as well.
            containers.append(containers[-1].superclass)
        return find_entry(node.attr, False, containers)
    elif isinstance(node, ast.Subscript):
        # This might be a type literal like List[int]
        if isinstance(node.value, ast.Name):
            module = next(cont for cont in containers
                          if isinstance(cont, PythonModule))
            type_class = None
            if node.value.id == 'Dict':
                type_class = module.global_module.classes[DICT_TYPE]
            if node.value.id == 'Set':
                type_class = module.global_module.classes[SET_TYPE]
            if node.value.id == 'List':
                type_class = module.global_module.classes[LIST_TYPE]
            if node.value.id == 'Tuple':
                type_class = module.global_module.classes[TUPLE_TYPE]
            if not type_class:
                possible_class = get_target(node.value, containers, container)
                if isinstance(possible_class, PythonType):
                    type_class = possible_class
            if type_class:
                # Look up the type arguments. Also consider string arguments.
                if isinstance(node.slice.value, ast.Tuple):
                    args = [get_target(arg, containers, container, True)
                            for arg in node.slice.value.elts]
                else:
                    args = [get_target(node.slice.value, containers, container, True)]
                return GenericType(type_class, args)
            if node.value.id == 'Optional':
                option = get_target(node.slice.value, containers, container, True)
                return OptionalType(option)
            if node.value.id == 'Union':
                if isinstance(node.slice.value, ast.Tuple):
                    elts = [get_target(e, containers, container, True)
                            for e in node.slice.value.elts]
                    return UnionType(elts)
                else:
                    return get_target(node.slice.value, containers, container, True)

    else:
        return None


def find_entry(target_name: str, only_top: bool,
               containers: List[ContainerInterface]) -> Optional[PythonNode]:
    """
    Returns the PythonNode identified by the given name in the given containers.
    """
    for lhs in containers:
        if lhs:
            options = lhs.get_contents(only_top=only_top)
            if target_name in options:
                return options[target_name]
    return None


def get_type(node: ast.AST, containers: List[ContainerInterface],
             container: PythonNode) -> Optional[PythonType]:
    """
    If ``node`` is an expression, returns its type, assuming that the immediate
    container (e.g. a PythonMethod) of the node is ``container``, by looking in
    the given ``containers`` (can be e.g. PythonMethods, the Context,
    PythonModules, etc). For primitive values, returns the boxed version.
    Returns None if the type is void.
    """
    result = _do_get_type(node, containers, container)
    if isinstance(result, PythonType):
        result = result.try_box()
    return result


def _do_get_type(node: ast.AST, containers: List[ContainerInterface],
                 container: PythonNode) -> Optional[PythonType]:
    """
    Does the actual work for get_type without boxing the type.
    """
    if isinstance(container, (PythonIOOperation, PythonMethod)):
        module = container.module
        current_function = container
    else:
        module = container
        current_function = None
    target = get_target(node, containers, container)
    if target:
        if isinstance(target, PythonVarBase):
            return target.get_specific_type(node)
        if isinstance(target, PythonMethod):
            if isinstance(node, ast.Call) and isinstance(node.func, ast.Attribute):
                rec_target = get_target(node.func.value, containers, container)
                if not isinstance(rec_target, PythonModule):
                    rectype = get_type(node.func.value, containers, container)
                    if target.generic_type != -1:
                        return rectype.type_args[target.generic_type]
                    if isinstance(target.type, TypeVar):
                        while rectype.python_class is not target.cls:
                            rectype = rectype.superclass
                        name_list = list(rectype.python_class.type_vars.keys())
                        index = name_list.index(target.type.name)
                        return rectype.type_args[index]
            return target.type
        if isinstance(target, PythonField):
            result = target.type
            if isinstance(result, TypeVar):
                assert isinstance(node, ast.Attribute)
                rec_type = _do_get_type(node.value, containers, container)
                while (rec_type.python_class is not
                        result.target_type.python_class):
                    rec_type = rec_type.superclass
                result = rec_type.type_args[result.index]
            return result
        if isinstance(target, PythonIOOperation):
            return module.global_module.classes[BOOL_TYPE]

        if isinstance(target, (PythonType, PythonModule)):
            if (isinstance(node, ast.Call) and
                    isinstance(target, PythonClass) and
                    target.type_vars):
                # This is a call to a constructor of a generic class; it's not
                # enough to just return the class, we need the entire type with
                # type arguments. We only support that if we can get it directly
                # from mypy, i.e., when the result is assigned to a variable
                # and we can get the variable type.
                if node._parent and isinstance(node._parent, ast.Assign):
                    return get_type(node._parent.targets[0], containers,
                                    container)
                elif (target.name in (PSEQ_TYPE, PSET_TYPE, PMSET_TYPE) and
                          isinstance(node, ast.Call) and node.args):
                    arg_types = [get_type(arg, containers, container)
                                 for arg in node.args]
                    return GenericType(target, [common_supertype(arg_types)])
                else:
                    error = 'generic.constructor.without.type'
                    raise InvalidProgramException(node, error)
            return target
    if isinstance(node, (ast.Attribute, ast.Name)):
        if isinstance(node, ast.Attribute):
            lhs = _do_get_type(node.value, containers, container)
            if isinstance(lhs, UnionType) and not isinstance(lhs, OptionalType):
                candidates = [find_entry(node.attr, False, [t]) for t in lhs.type_args]
                if all(isinstance(c, (PythonField, PythonVarBase)) for c in candidates):
                    return common_supertype([c.type for c in candidates])
        # All these cases should be handled by get_target, so if we get here,
        # the node refers to something unknown in the given context.
        return None
    if isinstance(node, ast.Num):
        return module.global_module.classes[INT_TYPE]
    elif isinstance(node, ast.Tuple):
        args = [get_type(arg, containers, container) for arg in node.elts]
        return GenericType(module.global_module.classes[TUPLE_TYPE],
                           args)
    elif isinstance(node, ast.Subscript):
        return get_subscript_type(node, module, containers, container)
    elif isinstance(node, ast.Str):
        return module.global_module.classes[STRING_TYPE]
    elif isinstance(node, ast.Bytes):
        return module.global_module.classes[BYTES_TYPE]
    elif isinstance(node, ast.Compare):
        return module.global_module.classes[BOOL_TYPE]
    elif isinstance(node, ast.BoolOp):
        # And and Or always return one of their operands, so we use the common
        # supertype of all arguments.
        # TODO: We could also use a union type, but since support for e.g.
        # calling methods on those isn't amazing yet, we don't do that yet.
        operand_types = [get_type(operand, containers, container)
                         for operand in node.values]
        return common_supertype(operand_types)
    elif isinstance(node, ast.List):
        return _get_collection_literal_type(node, ['elts'], LIST_TYPE, module,
                                            containers, container)
    elif isinstance(node, ast.Set):
        return _get_collection_literal_type(node, ['elts'], SET_TYPE, module,
                                            containers, container)
    elif isinstance(node, ast.Dict):
        return _get_collection_literal_type(node, ['keys', 'values'], DICT_TYPE,
                                            module, containers, container)
    elif isinstance(node, ast.IfExp):
        body_type = get_type(node.body, containers, container)
        else_type = get_type(node.orelse, containers, container)
        return pairwise_supertype(body_type, else_type)
    elif isinstance(node, ast.BinOp):
        left_type = get_type(node.left, containers, container)
        right_type = get_type(node.right, containers, container)
        operator_func = OPERATOR_FUNCTIONS[type(node.op)]
        return left_type.get_func_or_method(operator_func).type
    elif isinstance(node, ast.UnaryOp):
        if isinstance(node.op, ast.Not):
            return module.global_module.classes[BOOL_TYPE]
        elif isinstance(node.op, ast.USub):
            return module.global_module.classes[INT_TYPE]
        else:
            raise UnsupportedException(node)
    elif isinstance(node, ast.NameConstant):
        if (node.value is True) or (node.value is False):
            return module.global_module.classes[BOOL_TYPE]
        elif node.value is None:
            return module.global_module.classes[OBJECT_TYPE]
        else:
            raise UnsupportedException(node)
    elif isinstance(node, ast.Call):
        return _get_call_type(node, module, current_function, containers,
                              container)
    elif isinstance(node, ast.ListComp):
        if (node._parent and isinstance(node._parent, ast.Assign) and
                    node is node._parent.value):
            # Constructor is assigned to variable;
            # we get the type of the dict from the type of the
            # variable it's assigned to.
            return get_type(node._parent.targets[0], containers,
                            container)
        else:
            raise UnsupportedException(node, 'List comprehensions must be directly '
                                       'assigned to a local variable.')
    else:
        raise UnsupportedException(node)


def _get_collection_literal_type(node: ast.AST, arg_fields: List[str],
                                 coll_type: str, module: PythonModule,
                                 containers: List[ContainerInterface],
                                 container: PythonNode) -> PythonType:
    """
    Assuming ``node`` is a collection literal, ``coll_type`` is the type of the
    collection (e.g. 'list'), and ``arg_fields`` contains the fields of the
    literal which contain the contents of the literal (e.g. 'keys' and 'values'
    for a dict), returns the type of the collection.
    """
    if hasattr(node, '_parent') and isinstance(node._parent, ast.Assign):
        # Constructor is assigned to variable;
        # we get the type of the dict from the type of the
        # variable it's assigned to.
        args = get_type(node._parent.targets[0], containers,
                        container).type_args
    elif all(getattr(node, arg_field) for arg_field in arg_fields):
        args = []
        for arg_field in arg_fields:
            arg_types = [get_type(arg, containers, container) for arg in
                         getattr(node, arg_field)]
            args.append(common_supertype(arg_types))
    else:
        object_class = module.global_module.classes[OBJECT_TYPE]
        args = [object_class for arg_field in arg_fields]
    return GenericType(module.global_module.classes[coll_type],
                       args)


def _get_call_type(node: ast.Call, module: PythonModule,
                   current_function: PythonMethod,
                   containers: List[ContainerInterface],
                   container: PythonNode) -> PythonType:
    func_name = get_func_name(node)
    if func_name == 'super':
        if len(node.args) == 2:
            return module.classes[node.args[0].id].superclass
        elif not node.args:
            return container.cls.superclass
        else:
            raise InvalidProgramException(node, 'invalid.super.call')
    if func_name == 'len':
        return module.global_module.classes[INT_TYPE]
    if func_name in ('token', 'ctoken', 'MustTerminate', 'MustRelease'):
        return module.global_module.classes[BOOL_TYPE]
    if func_name == PSEQ_TYPE:
        return _get_collection_literal_type(node, ['args'], PSEQ_TYPE, module,
                                            containers, container)
    if func_name == PSET_TYPE:
        return _get_collection_literal_type(node, ['args'], PSET_TYPE, module,
                                            containers, container)
    if func_name == PMSET_TYPE:
        return _get_collection_literal_type(node, ['args'], PMSET_TYPE, module,
                                            containers, container)
    if func_name == 'enumerate':
        if len(node.args) != 1:
            raise UnsupportedException(node, 'enumerate only supported with single arg.')
        list_type = module.global_module.classes[LIST_TYPE]
        int_type = module.global_module.classes[INT_TYPE]
        tuple_type = module.global_module.classes[TUPLE_TYPE]
        arg_type = get_type(node.args[0], containers, container)
        iterable_type = _get_iteration_type(arg_type, module, node)
        return GenericType(list_type, [GenericType(tuple_type,
                                                   [int_type, iterable_type])])
    if isinstance(node.func, ast.Name):
        if node.func.id in CONTRACT_FUNCS:
            if node.func.id  == 'Result':
                return current_function.type
            elif node.func.id == 'RaisedException':
                ctxs = [cont for cont in containers if
                        hasattr(cont, 'var_aliases')]
                ctx = ctxs[0] if ctxs else None
                assert ctx
                assert ctx.current_contract_exception is not None
                return ctx.current_contract_exception
            elif node.func.id in ('Acc', 'Rd', 'Read', 'Implies', 'Forall', 'Exists',
                                  'MayCreate', 'MaySet', 'Low', 'LowVal', 'LowEvent'):
                return module.global_module.classes[BOOL_TYPE]
            elif node.func.id == 'Declassify':
                return None
            elif node.func.id == 'Old':
                return get_type(node.args[0], containers, container)
            elif node.func.id == 'Unfolding':
                return get_type(node.args[1], containers, container)
            elif node.func.id == 'ToSeq':
                arg_type = get_type(node.args[0], containers, container)
                seq_class = module.global_module.classes[PSEQ_TYPE]
                content_type = _get_iteration_type(arg_type, module, node)
                return GenericType(seq_class, [content_type])
            elif node.func.id == 'Previous':
                arg_type = get_type(node.args[0], containers, container)
                list_class = module.global_module.classes[PSEQ_TYPE]
                return GenericType(list_class, [arg_type])
            elif node.func.id in ('getArg', 'getOld', 'getMethod'):
                object_class = module.global_module.classes[OBJECT_TYPE]
                return object_class
            elif node.func.id == 'Let':
                body_type = get_target(node.args[1], containers, container)
                if isinstance(body_type, PythonType):
                    return body_type
                raise InvalidProgramException(node, 'invalid.let')
            else:
                raise UnsupportedException(node)
        elif node.func.id in BUILTINS:
            if node.func.id in ('isinstance', BOOL_TYPE):
                return module.global_module.classes[BOOL_TYPE]
            elif node.func.id == 'cast':
                return get_target(node.args[0],
                                  containers, container)
            else:
                raise UnsupportedException(node)
        if node.func.id in module.classes:
            return module.global_module.classes[node.func.id]
        elif module.get_func_or_method(node.func.id) is not None:
            target = module.get_func_or_method(node.func.id)
            return target.type
    elif isinstance(node.func, ast.Attribute):
        rectype = get_type(node.func.value, containers, container)
        if isinstance(rectype, UnionType):
            set_of_classes = rectype.get_types() - {None}
            set_of_return_types = {type.get_func_or_method(node.func.attr).type
                                   for type in set_of_classes}
            if len(set_of_return_types) == 1:
                return set_of_return_types.pop()
            elif len(set_of_return_types) == 2 and None in set_of_return_types:
                return OptionalType((set_of_return_types - {None}).pop())
            else:
                return UnionType(list(set_of_return_types))
        elif isinstance(rectype, PythonType):
            target = rectype.get_func_or_method(node.func.attr)
            if target.generic_type != -1:
                return rectype.type_args[target.generic_type]
            else:
                return target.type
    else:
        raise UnsupportedException(node)


def get_subscript_type(node: ast.Subscript, module: PythonModule,
                        containers: List[ContainerInterface],
                        container: PythonNode) -> PythonType:
    if (node._parent and isinstance(node._parent, ast.Assign) and
            node is node._parent.value):
        # Constructor is assigned to variable;
        # we get the type of the dict from the type of the
        # variable it's assigned to.
        return get_type(node._parent.targets[0], containers,
                        container)
    value_type = get_type(node.value, containers, container)
    return _get_subscript_type(value_type, module, node)


def _get_subscript_type(value_type: PythonType, module: PythonModule,
                        node: ast.AST) -> PythonType:
    if isinstance(value_type, OptionalType):
        value_type = value_type.cls
    if value_type.name == TUPLE_TYPE:
        if isinstance(node, ast.Subscript):
            if isinstance(node.slice, ast.Slice):
                raise UnsupportedException(node, 'tuple slicing')
            if len(value_type.type_args) == 1:
                return value_type.type_args[0]
            if isinstance(node.slice.value, ast.UnaryOp):
                if (isinstance(node.slice.value.op, ast.USub) and
                        isinstance(node.slice.value.operand, ast.Num)):
                    index = -node.slice.value.operand.n
                else:
                    raise UnsupportedException(node, 'dynamic subscript type')
            elif isinstance(node.slice.value, ast.Num):
                index = node.slice.value.n
            return value_type.type_args[index]
        else:
            return common_supertype(value_type.type_args)
    elif value_type.name == LIST_TYPE:
        return value_type.type_args[0]
    elif value_type.name == SET_TYPE:
        return value_type.type_args[0]
    elif value_type.name in (DICT_TYPE, 'defaultdict', 'ExpiringDict'):
        # FIXME: This is very unfortunate, but right now we cannot handle this
        # generically, so we have to hard code these two cases for the moment.
        return value_type.type_args[1]
    elif value_type.name in (RANGE_TYPE, BYTES_TYPE):
        return module.global_module.classes[INT_TYPE]
    elif value_type.name == PSEQ_TYPE:
        return value_type.type_args[0]
    elif value_type.name == PSET_TYPE:
        return value_type.type_args[0]
    elif value_type.name == PMSET_TYPE:
        return value_type.type_args[0]
    elif value_type.python_class.get_func_or_method('__getitem__'):
        return value_type.python_class.get_func_or_method('__getitem__').type
    else:
        raise UnsupportedException(node)


def _get_iteration_type(value_type: PythonType, module: PythonModule,
                        node: ast.AST) -> PythonType:
    # Assuming that value_type is some sort of container which supports iteration,
    # returns the type of the iterator.
    if value_type.name in (DICT_TYPE, 'defaultdict', 'ExpiringDict'):
        # FIXME: This is very unfortunate, but right now we cannot handle this
        # generically, so we have to hard code these two cases for the moment.
        return value_type.type_args[0]
    else:
        return _get_subscript_type(value_type, module, node)


def common_supertype(types: List[PythonType]) -> Optional[PythonType]:
    """
    Returns the common supertype of all types in the list. The list may not
    be empty.
    """
    assert types
    if len(types) == 1:
        return types[0]
    current = types[0]
    for new in types[1:]:
        current = pairwise_supertype(current, new)
    return current


def pairwise_supertype(t1: PythonType, t2: PythonType) -> Optional[PythonType]:
    """
    Returns the common supertype of 't1' and 't2', if any.
    """
    if t1.issubtype(t2):
        return t2
    if t2.issubtype(t1):
        return t1
    if (not t1.superclass and not t2.superclass):
        return None
    if not t1.superclass:
        return pairwise_supertype(t2.superclass, t1)
    return pairwise_supertype(t2, t1.superclass)


"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Common code for obligation translators."""


import abc
import ast

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.typedefs import (
    StmtsAndExpr,
)
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.util import (
    join_expressions,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.abstract import TranslatorConfig
from nagini_translation.translators.common import CommonTranslator
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)
from nagini_translation.translators.obligation.obligation_info import (
    BaseObligationInfo,
)
from nagini_translation.translators.obligation.types.base import (
    ObligationInstance,
)
from nagini_translation.translators.obligation.types.must_invoke import (
    MustInvokeObligationInstance,
)
from nagini_translation.translators.obligation.types.must_release import (
    MustReleaseObligationInstance,
)
from nagini_translation.translators.obligation.types.must_terminate import (
    MustTerminateObligationInstance,
)


class CommonObligationTranslator(CommonTranslator):
    """Base class for obligation translators."""

    def __init__(self, config: TranslatorConfig, jvm: JVM, source_file: str,
                 type_info: TypeInfo, viper_ast: ViperAST,
                 obligation_manager: ObligationManager) -> None:
        super().__init__(config, jvm, source_file, type_info, viper_ast)
        self._obligation_manager = obligation_manager

    @abc.abstractmethod
    def _get_obligation_info(self, ctx: Context) -> BaseObligationInfo:
        """Get the relevant obligation info."""

    @abc.abstractmethod
    def _create_obligation_instance_use(
            self, obligation_instance: ObligationInstance,
            ctx: Context) -> sil.InhaleExhale:
        """Create obligation use from the obligation instance."""

    def _translate_obligation_use(
            self, node: ast.Call, ctx: Context,
            expected_type: type) -> StmtsAndExpr:
        obligation_info = self._get_obligation_info(ctx)
        guarded_obligation_instance = obligation_info.get_instance(node)
        obligation_instance = guarded_obligation_instance.obligation_instance
        assert isinstance(obligation_instance, expected_type)

        exprs_with_rules = self._create_obligation_instance_use(
            obligation_instance, ctx)

        info = self.no_info(ctx)
        translated_expressions = []
        for expression, rules in exprs_with_rules:
            position = self.to_position(node, ctx, rules=rules)
            translated_expression = expression.translate(
                self, ctx, position, info)
            translated_expressions.append(translated_expression)
        and_operator = (
            lambda left, right:
            self.viper.And(left, right, position, info))
        return ([], join_expressions(and_operator, translated_expressions))

    def translate_must_invoke(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate ``token`` in a contract."""
        return self._translate_obligation_use(
            node, ctx, MustInvokeObligationInstance)

    def translate_may_invoke(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate ``ctoken`` in a contract."""
        obligation = self._obligation_manager.must_invoke_obligation
        use = obligation.create_ctoken_use(node)
        position = self.to_position(node, ctx)
        return ([], use.translate(self, ctx, position, self.no_info(ctx)))

    def translate_must_release(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate ``MustRelease`` in a contract."""
        return self._translate_obligation_use(
            node, ctx, MustReleaseObligationInstance)

    def translate_must_terminate(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate ``MustTerminate`` in a contract."""
        return self._translate_obligation_use(
            node, ctx, MustTerminateObligationInstance)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Function call Silver expression."""

from typing import List

from nagini_translation.lib.silver_nodes.expression import Expression
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)


class CallArg:
    """A call argument."""

    def __init__(
            self, parameter_name: str, parameter_type: 'Type',
            argument: Expression) -> None:
        self._parameter_name = parameter_name
        self._parameter_type = parameter_type
        self._argument = argument

    def construct_definition(
            self, translator: 'Translator', position: Position,
            info: Info) -> Expr:
        """Construct formal argument definition."""
        typ = self._parameter_type.translate(translator)
        return translator.viper.LocalVarDecl(
            self._parameter_name, typ, position, info)

    def translate_argument(
            self, translator: 'AbstractTranslator', ctx: 'Context',
            position: Position, info: Info) -> Expr:
        """Translate the argument passed to the function."""
        result = self._argument.translate(translator, ctx, position, info)
        result = self._parameter_type.adjust_type(translator, result, ctx)
        return result


class CallMixin(Expression):
    """Generic function call mix-in.

    This mix-in should be mixed with an expression type to get a call to
    a function of that type.
    """

    def __init__(self, function: str, args: List[CallArg],
                 typ: 'Type') -> None:
        self._function = function
        self._args = args
        self._type = typ

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        formal_args = [
            arg.construct_definition(translator, position, info)
            for arg in self._args]
        args = [
            arg.translate_argument(translator, ctx, position, info)
            for arg in self._args]
        typ = self._type.translate(translator)
        return translator.viper.FuncApp(
            self._function, args, position, info, typ, formal_args)


__all__ = ('CallArg',)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""
This module contains code responsible for translating IO operations.

Translation of IO Existential Variables
=======================================

VeriFast has ``?a`` syntax which is essentially an assignment expression
that allows to link IO operations in contracts. However, neither Python,
nor Silver has assignment expressions.

``IOExists`` is a special construct that allows to define IO existential
variables (class ``PythonIOExistentialVar``) that can be used for
linking IO operations in contracts. For example::

    def read_int(t1: Place) -> Tuple[Place, int]:
        IOExists = lambda t2, value: (
            Requires(
                token(t1) and
                read_int_io(t1, value, t2)
            ),
            Ensures(
                token(t2) and
                t2 == Result()[0] and
                value == Result()[1]
            )
        )   # type: Callable[[Place, int], Tuple[bool, bool]]

Here ``t2`` and ``value`` are IO existential variables. Unlike normal
variables, existential variables are not created as variables on the
Silver level, but instead they are replaced with their definitions. A
definition of the existential variable is its first mention in a
contract, which must be one of:

1.  **IO operation's result.** In this case the definition of the
    existential variable is IO operation's getter. For example,
    ``read_int_io(t1, value, t2)`` in the example above defines
    ``value`` and ``t2``. As a result, in all subsequent uses
    ``value`` is translated to ``get__read_int_io__value(t1)`` and
    ``t2`` to ``get__read_int_io__t_post(t1)``.
2.  **Equality with already defined value.** The only accepted syntax in
    this case is ``existential_variable == something``. For example,
    ``2 == value`` would give an error because existential variable is
    on the right hand side. In this case, the definition of the
    existential variable is the right hand side of the equality.

    .. note::

        The defining equality must be a top level assertion because the
        following contract::

            (
                value == x.f
                if b
                else value == x.g
            ) and
            value == 2

        would be translated to:

        .. code-block:: silver

            (b ? True : x.f == x.g) && x.f == 2

        which is probably not what a programmer intended.

------------
Known Issues
------------

Heap Dependent Getters
----------------------

If one of the IO operation arguments is a field, then the emitted
getters are heap dependent. For example, the value of place ``t2``
depends on field ``self.int_field``::

    write_int_io(t1, self.int_field, t2)

This has interesting consequences such as:

1.  Postcondition must have access to ``self.int_field``, otherwise
    ``t2`` getter is not framed::

        IOExists1(Place)(
            lambda t2: (
            Requires(
                Acc(self.int_field) and
                write_int_io(t1, self.int_field, t2)
            ),
            Ensures(
                t2 == Result() # ERROR: not.wellformed:insufficient.permission
            ),
            )
        )

2.  Similarly, if defining getter is heap dependent and guarded by
    conditional, other branch fails well-formedness check::

        Requires(
            token(t1) and
            (
                Acc(self.int_field1, 1/2) and
                write_int_io(t1, self.int_field1, t2)
            ) if b else (
                Acc(self.int_field2, 1/2) and
                write_int_io(t1, self.int_field2, t2)
                               # ERROR: not.wellformed:insufficient.permission
            )
        ),

3.  If defining getter is changed from heap independent in overridden
    method to heap dependent in a overriding method and overriding
    method takes all permission to the heap location, the behavioural
    subtyping check fails because information about getter equality is
    havocked.

Currently, the plan is to ignore the problem because storing arguments
in fields should not be too common in practise:

1.  The Petri Net provided at the entry point cannot depend on the heap
    – otherwise also permissions has to be provided at the entry
    point, which does not make much sense.
2.  It is not allowed to have permissions in non-basic IO operation
    definitions.

.. todo:: Vytautas

    Things to investigate:

    1.  Does wrapping getters in ``old`` in postcondition solve the
        issue of having to provide permissions to fields in
        postcondition?
"""


from nagini_translation.translators.io_operation.interface import (
    IOOperationTranslator,
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from abc import ABCMeta
from collections import OrderedDict
from enum import Enum
from nagini_translation.lib import config
from nagini_translation.lib.errors import error_manager
from nagini_translation.lib.jvmaccess import JVM
from typing import Optional


class ViperVerifier(Enum):
    silicon = 'silicon'
    carbon = 'carbon'


class VerificationResult(metaclass=ABCMeta):
    pass


class Success(VerificationResult):
    """
    Encodes a verification success
    """

    def __bool__(self):
        return True

    def to_string(self, ide_mode: bool, show_viper_errors: bool) -> str:
        return "Verification successful"


class Failure(VerificationResult):
    """
    Encodes a verification failure and provides access to the errors
    """

    def __init__(
            self, errors: 'silver.verifier.AbstractError',
            jvm: Optional[JVM] = None):
        self.errors = error_manager.convert(errors, jvm)

    def __bool__(self):
        return False

    def to_string(self, ide_mode: bool, show_viper_errors: bool) -> str:
        all_errors = [error.string(ide_mode, show_viper_errors) for error in self.errors]
        unique_errors = []
        for e in all_errors:
            if e not in unique_errors:
                unique_errors.append(e)
        return "Verification failed\nErrors:\n" + '\n'.join(unique_errors)


class ARPPlugin:
    """
    Provides access to the ARPPlugin
    """

    def __init__(self, jvm: JVM):
        self.jvm = jvm
        self.silver = jvm.viper.silver
        if not jvm.is_known_class(jvm.viper.silver.plugin.ARPPlugin):
            raise Exception('ARP plugin not found on classpath.')
        self.arpplugin = jvm.viper.silver.plugin.ARPPlugin()
        self.set_ignored_fields()

    def before_verify(self, prog: 'silver.ast.Program') -> 'silver.ast.Program':
        return self.arpplugin.beforeVerify(prog)

    def map_result(self, result: 'verifier.VerificationResult') -> 'verifier.VerificationResult':
        return self.arpplugin.mapVerificationResult(result)

    def set_ignored_fields(self) -> None:
        fields = self.jvm.scala.collection.mutable.ArraySeq(6)
        fields.update(0, "MustReleaseBounded")
        fields.update(1, "MustReleaseUnbounded")
        fields.update(2, "MustTerminate")
        fields.update(3, "MustInvokeBounded")
        fields.update(4, "MustInvokeUnbounded")
        fields.update(5, "MustInvokeCredit")
        self.arpplugin.setIgnoredFields(fields)


_ARP_PLUGIN = None


def get_arp_plugin(jvm: JVM) -> ARPPlugin:
    global _ARP_PLUGIN
    if not _ARP_PLUGIN:
        _ARP_PLUGIN = ARPPlugin(jvm)
    return _ARP_PLUGIN


class Silicon:
    """
    Provides access to the Silicon verifier
    """

    def __init__(self, jvm: JVM, filename: str):
        self.jvm = jvm
        self.silver = jvm.viper.silver
        if not jvm.is_known_class(jvm.viper.silicon.Silicon):
            raise Exception('Silicon backend not found on classpath.')
        self.silicon = jvm.viper.silicon.Silicon()
        args = jvm.scala.collection.mutable.ArraySeq(4)
        args.update(0, '--z3Exe')
        args.update(1, config.z3_path)
        args.update(2, '--disableCatchingExceptions')
        args.update(3, filename)
        self.silicon.parseCommandLine(args)
        self.silicon.start()
        self.ready = True

    def verify(self, prog: 'silver.ast.Program', arp=False) -> VerificationResult:
        """
        Verifies the given program using Silicon
        """
        if not self.ready:
            self.silicon.restart()
        result = self.silicon.verify(prog)
        if arp:
            result = get_arp_plugin(self.jvm).map_result(result)
        self.ready = False
        if isinstance(result, self.silver.verifier.Failure):
            it = result.errors().toIterator()
            errors = []
            while it.hasNext():
                errors += [it.next()]
            return Failure(errors, self.jvm)
        else:
            return Success()

    def __del__(self):
        if hasattr(self, 'silicon') and self.silicon:
            self.silicon.stop()


class Carbon:
    """
    Provides access to the Carbon verifier
    """

    def __init__(self, jvm: JVM, filename: str):
        self.silver = jvm.viper.silver
        if not jvm.is_known_class(jvm.viper.carbon.CarbonVerifier):
            raise Exception('Carbon backend not found on classpath.')
        self.carbon = jvm.viper.carbon.CarbonVerifier()
        args = jvm.scala.collection.mutable.ArraySeq(5)
        args.update(0, '--boogieExe')
        args.update(1, config.boogie_path)
        args.update(2, '--z3Exe')
        args.update(3, config.z3_path)
        args.update(4, filename)
        self.carbon.parseCommandLine(args)
        self.carbon.start()
        self.ready = True
        self.jvm = jvm

    def verify(self, prog: 'silver.ast.Program', arp=False) -> VerificationResult:
        """
        Verifies the given program using Carbon
        """
        if not self.ready:
            self.carbon.restart()
        result = self.carbon.verify(prog)
        if arp:
            result = get_arp_plugin(self.jvm).map_result(result)
        self.ready = False
        if isinstance(result, self.silver.verifier.Failure):
            it = result.errors().toIterator()
            errors = []
            while it.hasNext():
                errors += [it.next()]
            return Failure(errors)
        else:
            return Success()
